#ifndef __sun_h__
#define __sun_h__

#include <arpa/inet.h>
#include <time.h>
#include <assert.h>
#include <stdio.h>
#include <openssl/sha.h>

#include "parameters.h"
#include "config.h"

#define panic(a...) do { fprintf(stderr, a); fprintf(stderr, "\n"); exit(1); } while (0)

#define INFINITE_RTT 99999999

#define BOOTSTRAP_SUN_INTERVAL_SEC 1
#define ONLINE_SUN_INTERVAL_SEC 59
#define BOOTSTRAP_PLANET_INTERVAL_SEC 1
#define ONLINE_PLANET_INTERVAL_SEC 1
#define BOOTSTRAP_TIMEOUT_PER_PLANET_SEC 3
#define MAX_BOOTSTRAP_ATTEMPTS 2

#define EDATABASE 1       /* database error */
#define ENOEXP 2          /* experiment not found */
#define EEXPIRED 3        /* experiment has expired */

class Log {
protected:
  FILE *file;
  bool logFlush;
  int loglevel;

  Log(FILE *file, int loglevel);
  
public:
  static Log *open(const char *filename, int loglevel, bool append);
  static Log *openEmpty();
  void close();
  ~Log();
  void setFlush(bool doFlush) { logFlush = doFlush; };
  void dump(int level, unsigned char *data, int len);
  void write(int level, const char *format, ...);
};

class SelectorCallback {
public:
  SelectorCallback() {};
  virtual void descriptorReady(int fd, bool canRead, bool canWrite, bool hasException) {};
  virtual void timerExpired(int id, int arg) {};
};

class Selector {
protected:
  const static int MAX_DESCRIPTORS = 100;
  const static int MAX_TIMERS = 100;

  struct descriptor {
    int fd;
    bool wantRead;
    bool wantWrite;
    bool wantExc;
    SelectorCallback *callback;
  } descriptors[MAX_DESCRIPTORS];

  struct timer {
    int id;
    int arg;
    long long whenMicros;
    SelectorCallback *callback;
  } timers[MAX_TIMERS];

  int numDescriptors;
  int numTimers;

public:
  Selector();
  void run();
  long long getTimeMicros();
  void setDescriptorCallback(int fd, bool wantRead, bool wantWrite, bool wantExceptions, SelectorCallback *callback);
  void cancelDescriptorCallback(int fd, SelectorCallback *callback);
  void setTimer(int id, int arg, long long whenMicros, SelectorCallback *callback);
  void cancelTimer(int id, int arg, SelectorCallback *callback);
};

class Database {
public:
  Database() {};
  virtual void getSatelliteAllocation(const char *email, const char *identifier, const int nowTS,
    bool *isAllocated, char *expOwner, char *expIdentifier, unsigned char *planetPwd,
    int *planetPwdLen, char *planetName, in_addr_t *planetIP, int *planetPort, in_addr_t *lastSatelliteIP, bool *allowMobility, int *nextRefreshTS
  ) = 0;
  virtual void recordSatelliteCredits(const char *email, const char *identifier, int fromTS,
    int onlineSec, in_addr_t ip, int port, const char *domain, int version, long long bytesSent,
    long long bytesReceived
  ) = 0;
  virtual bool statExperiment(const char *owner, const char *identifier, int *startTime, int *endTime,
    bool *dynAllocPlanets, bool *dynAllocSats, bool *allowSatMigration, int *satsLeftToAllocate,
    char *description, int descriptionMaxLen
  ) = 0;
  virtual bool statPlanet(const char *expOwner, const char *expIdentifier, in_addr_t ip, int port, 
    char *planetAddress
  ) = 0;
  virtual void addPlanet(const char *expOwner, const char *expIdentifier, const char *address, 
    in_addr_t ip, int port
  ) = 0;
  virtual int getNumSatellitesInExperiment(const char *expOwner, const char *expIdentifier) = 0;
  virtual int getNumPlanetsInExperiment(const char *expOwner, const char *expIdentifier) = 0;
  virtual int getExperiments(int fromPeriod, int toPeriod, char *buf, int bufsize, int maxEntries, char ** const owners, char ** const identifiers) = 0;
  virtual bool getMatrixForExperiment(const char *owner, const char *identifier, unsigned char *matrix, int matrixSize) = 0;
  virtual int getPlanets(const char *expOwner, const char *expIdentifier, char *buf, int bufsize, int maxEntries, char ** const names, char ** const ips, char ** const ports) = 0;
  virtual bool addSatellite(const char *expOwner, const char *expIdentifier, 
    const char *satEmail, const char *satIdentifier, const char *planetPassword, const char *virtualIP
  ) = 0;
  virtual bool addBinding(const char *expOwner, const char *expIdentifier, 
    const char *satEmail, const char *satIdentifier, const char *satIP, 
    const char *planetAddress, int planetPort
  ) = 0;
  virtual bool findBinding(const char *expOwner, const char *expIdentifier,
    const char *satEmail, const char *satIdentifier, in_addr_t satIP, 
    char *planetAddress, in_addr_t *planetIP, int *planetPort, int *validFrom
  ) = 0;
  virtual bool statUser(const char *username, char *institution, int maxInstitutionLen, char *password, int maxPasswordLen) = 0;
};

class Sun : public SelectorCallback {
protected:
  const static int MAX_SATELLITES = 500;
  const static int MAX_PLANETS = 100;
  const static int MAX_EMAIL_LEN = 100;
  const static int MAX_EXPERIMENTS = 100;
  const static int MAX_MATRIX_CACHE_ENTRIES = 100;
  const static int MAX_IDENTIFIER_LEN = 100;
  const static long long HOUSEKEEPING_INTERVAL_MICROS = 300*1000000LL;
  const static long long FLUSH_CREDIT_INTERVAL_MICROS = 3600*1000000LL;
  const static long long EXPERIMENT_REFRESH_INTERVAL_MICROS = 60*1000000LL;
  const static int TI_HOUSEKEEPING = 1;
  const static int TI_FLUSH_CREDIT = 2;
  const static int TI_EXPERIMENT_REFRESH = 3;
  const static int NUM_BOOTSTRAP_ATTEMPTS_PER_NODE = 2;
  const static int BOOTSTRAP_TIMEOUT_SEC_PER_ATTEMPT = 3;

public:
  const static int SATELLITE_TIMEOUT_SEC = 90;
  const static int PLANET_TIMEOUT_SEC = 90;
  const static int SATELLITE_CLEANUP_SEC = 300;
  const static int PLANET_CLEANUP_SEC = 300;
  const static int MATRIX_ENTRY_BYTES = 18;

  struct satelliteStruct {
    char email[MAX_EMAIL_LEN+1];
    char identifier[MAX_IDENTIFIER_LEN+1];
    char domain[100];
    unsigned int reportedBytesSent;
    unsigned int reportedBytesReceived;
    in_addr_t ip;
    int port;
    int version;
    unsigned char sunPassword[8];
    int lastWentOnline;
    int lastStatusReport;

    // Billing
    long long totalBytesSent;
    long long totalBytesReceived;
    
    // Allocation
    bool isAllocated;
    char expOwner[200];
    char expIdentifier[200];
    unsigned char planetPassword[20];
    int planetPasswordLen;
    char planetName[200];
    in_addr_t planetIP;
    int planetPort;
    int nextRefreshAllocation;
    
    // Bootstrap
    bool isBootstrapping;
    int bootstrapID;
    int currentBootstrapIndex;
    int lastChangedBootstrap;
    int numBootstraps;
    in_addr_t *bootIP;
    int *bootPort;
    char **bootAddress;
    int *bootAttempts;
    int *bootSuccesses;
    int *bootBestRTT;
  };
  
  struct planetStruct {
    in_addr_t ip;
    int port;
    int version;
    bool acceptsUpdates;
    unsigned char lastConfigHash[SHA_DIGEST_LENGTH];
    char address[200];
    char expOwner[200];
    char expIdentifier[200];
    int lastStatusReport;
  };
  
  struct experimentStruct {
    char owner[200];
    char identifier[200];
    char description[1000];
    int numSatellitesAllocated;
    int numSatellitesMissing;
    int numPlanetsAllocated;
    bool dynAllocSats;
    bool dynAllocPlanets;
    bool allowSatMigration;
    int startTime;
    int endTime;
  };
  
  struct matrixCacheEntry {
    char owner[200];
    char identifier[200];
    int lastQueried;
    unsigned char *matrix;
    int matrixSizeBytes;
    unsigned char matrixHash[SHA_DIGEST_LENGTH];
  };
  
protected:
  struct satelliteStruct satellite[MAX_SATELLITES];
  struct planetStruct planet[MAX_PLANETS];
  struct matrixCacheEntry matrixCache[MAX_MATRIX_CACHE_ENTRIES];
  struct experimentStruct experiment[MAX_EXPERIMENTS];
  Selector *selector;
  int masterSocket;
  int masterPort;
  int numMatrixCacheEntries;
  int numSatellites;
  int numPlanets;
  int numExperiments;
  int nextBootstrapID;
  Database *db;
  Log *log;
  
  void handleStatusReport(int masterSocket, in_addr_t ip, int port, unsigned char *data, int length);
  void handlePlanetReport(int masterSocket, in_addr_t ip, int port, unsigned char *data, int length);
  void handleRTTReport(int masterSocket, in_addr_t ip, int port, int rttMillis, unsigned char *planetPassword, int planetPasswordLen);
  void sendConfigToSatellite(int masterSocket, int idx);
  void getGoalSettingsForSatellite(int idx, int *sunInterval, int *planetInterval, in_addr_t *planetIp, int *planetPort, unsigned char *pwdBuf, unsigned int pwdMaxLen, int *pwdLen);
  void flushSatelliteEntry(int idx);
  void flushPlanetEntry(int idx);
  int findInMatrixCache(const char *owner, const char *identifier);
  void tryToAllocateSatellites();
  void startBootstrappingSatellite(int idx);
  void cancelBootstrappingSatellite(int idx);
  void finishBootstrappingSatellite(int idx);
  int getBestNextBootstrap(int idx);
  int findExperimentEntry(const char *owner, const char *identifier);
  void invalidateMatrix(const char *owner, const char *identifier);

public:
  const static int DEFAULT_MASTER_PORT = 30000;

  Sun(ParameterFile *params, Selector *selector, Log *log, Database *db);
  virtual void descriptorReady(int fd, bool canRead, bool canWrite, bool hasException);
  virtual void timerExpired(int id, int arg);
  void init();
  int numSatelliteRecords() { return numSatellites; };
  int numPlanetRecords() { return numPlanets; };
  int numExperimentRecords() { return numExperiments; };
  bool statSatelliteRecord(int idx, struct satelliteStruct *buf) { if ((idx<0) || (idx>=numSatellites)) return false; *buf = satellite[idx]; return true; };
  bool statPlanetRecord(int idx, struct planetStruct *buf) { if ((idx<0) || (idx>=numPlanets)) return false; *buf = planet[idx]; return true; };
  bool statExperimentRecord(int idx, struct experimentStruct *buf) { if ((idx<0) || (idx>=numExperiments)) return false; *buf = experiment[idx]; return true; };
  bool isPlanetConfiguredCorrectly(int idx);
  bool isValidAuthentication(const char *owner, const char *auth);
  void reassociateSatellites(const char *owner, const char *identifier);
};

class WebFrontend : public SelectorCallback {
protected:
  const static int DEFAULT_HTTP_PORT = 8000;
  const static int MAX_CONNECTIONS = 10;

  struct connectionStruct {
    bool active;
    int fd;
    bool isReading;
    char buf[20000];
    int ptr;
    int len;
  } connection[MAX_CONNECTIONS]; 

  char hostname[200];
  char adminAuth[200];
  Selector *selector;
  Sun *sun;
  Log *log;
  int httpPort;
  int httpSocket;
  
  void maybeHandleHttpRequest(int index);
  void authenticationFailed(int index, const char *authDomain);
  
public:
  WebFrontend(ParameterFile *params, Selector *selector, Log *log, Sun *sun);
  virtual void descriptorReady(int fd, bool canRead, bool canWrite, bool hasException);
  void init();
};

class MysqlDatabase : public Database {
protected:
  char dbName[200];
  char dbUsername[200];
  char dbPassword[200];
  Selector *selector;
  Log *log;

  int mysqlQuery(const char *dbName, const char *username, const char *passwd, const char *query, char *buf, int maxlen, char **arg, int maxColumns, int maxEntries);

public:
  MysqlDatabase(ParameterFile *params, Selector *selector, Log *log);

  virtual void getSatelliteAllocation(const char *email, const char *identifier, const int nowTS,
    bool *isAllocated, char *expOwner, char *expIdentifier, unsigned char *planetPwd,
    int *planetPwdLen, char *planetName, in_addr_t *planetIP, int *planetPort, 
    in_addr_t *lastSatelliteIP, bool *allowMobility, int *nextRefreshTS
  );
  virtual void recordSatelliteCredits(const char *email, const char *identifier, int fromTS,
    int onlineSec, in_addr_t ip, int port, const char *domain, int version, long long bytesSent,
    long long bytesReceived
  );
  virtual bool statExperiment(const char *owner, const char *identifier, int *startTime, int *endTime,
    bool *dynAllocPlanets, bool *dynAllocSats, bool *allowSatMigration, int *satsLeftToAllocate,
    char *description, int descriptionMaxLen
  );
  virtual bool statPlanet(const char *expOwner, const char *expIdentifier, in_addr_t ip, int port, 
    char *planetAddress
  );
  virtual void addPlanet(const char *expOwner, const char *expIdentifier, const char *address, 
    in_addr_t ip, int port
  );
  virtual int getNumSatellitesInExperiment(const char *expOwner, const char *expIdentifier);
  virtual bool getMatrixForExperiment(const char *owner, const char *identifier, unsigned char *matrix, int matrixSize);
  virtual int getNumPlanetsInExperiment(const char *expOwner, const char *expIdentifier);
  virtual int getExperiments(int fromPeriod, int toPeriod, char *buf, int bufsize, int maxEntries, char **owners, char **identifiers);
  virtual int getPlanets(const char *expOwner, const char *expIdentifier, char *buf, int bufsize, int maxEntries, char ** const names, char ** const ips, char ** const ports);
  virtual bool addSatellite(const char *expOwner, const char *expIdentifier, 
    const char *satEmail, const char *satIdentifier, const char *planetPassword, const char *virtualIP
  );
  virtual bool addBinding(const char *expOwner, const char *expIdentifier, 
    const char *satEmail, const char *satIdentifier, const char *satIP, 
    const char *planetAddress, int planetPort
  );
  virtual bool findBinding(const char *expOwner, const char *expIdentifier,
    const char *satEmail, const char *satIdentifier, in_addr_t satIP, 
    char *planetAddress, in_addr_t *planetIP, int *planetPort, int *validFrom
  );
  virtual bool statUser(const char *username, char *institution, int maxInstitutionLen, char *password, int maxPasswordLen);
};

class FileDatabase : public Database {
protected:
  struct planetStruct {
    char name[200];
    in_addr_t ip;
    char sip[20];
    int port;
    char sport[10];
  };

  struct satelliteStruct {
    char email[200];
    char identifier[200];
    int planetIndex;
  };
    
  struct satelliteStruct *satellite;
  struct planetStruct *planet;
  char inputFile[200];
  char theIdentifier[200];
  char theOwner[200];
  int numSatellites;
  int numPlanets;
  int startTime;
  Selector *selector;
  Log *log;

public:
  FileDatabase(ParameterFile *params, Selector *selector, Log *log);

  virtual void getSatelliteAllocation(const char *email, const char *identifier, const int nowTS,
    bool *isAllocated, char *expOwner, char *expIdentifier, unsigned char *planetPwd,
    int *planetPwdLen, char *planetName, in_addr_t *planetIP, int *planetPort, 
    in_addr_t *lastSatelliteIP, bool *allowMobility, int *nextRefreshTS
  );
  virtual void recordSatelliteCredits(const char *email, const char *identifier, int fromTS,
    int onlineSec, in_addr_t ip, int port, const char *domain, int version, long long bytesSent,
    long long bytesReceived
  );
  virtual bool statExperiment(const char *owner, const char *identifier, int *startTime, int *endTime,
    bool *dynAllocPlanets, bool *dynAllocSats, bool *allowSatMigration, int *satsLeftToAllocate,
    char *description, int descriptionMaxLen
  );
  virtual bool statPlanet(const char *expOwner, const char *expIdentifier, in_addr_t ip, int port, 
    char *planetAddress
  );
  virtual void addPlanet(const char *expOwner, const char *expIdentifier, const char *address, 
    in_addr_t ip, int port
  );
  virtual int getNumSatellitesInExperiment(const char *expOwner, const char *expIdentifier);
  virtual bool getMatrixForExperiment(const char *owner, const char *identifier, unsigned char *matrix, int matrixSize);
  virtual int getNumPlanetsInExperiment(const char *expOwner, const char *expIdentifier);
  virtual int getExperiments(int fromPeriod, int toPeriod, char *buf, int bufsize, int maxEntries, char ** const owners, char ** const identifiers);
  virtual int getPlanets(const char *expOwner, const char *expIdentifier, char *buf, int bufsize, int maxEntries, char ** const names, char ** const ips, char ** const ports);
  virtual bool addSatellite(const char *expOwner, const char *expIdentifier, 
    const char *satEmail, const char *satIdentifier, const char *planetPassword, const char *virtualIP
  );
  virtual bool addBinding(const char *expOwner, const char *expIdentifier, 
    const char *satEmail, const char *satIdentifier, const char *satIP, 
    const char *planetAddress, int planetPort
  );
  virtual bool findBinding(const char *expOwner, const char *expIdentifier,
    const char *satEmail, const char *satIdentifier, in_addr_t satIP, 
    char *planetAddress, in_addr_t *planetIP, int *planetPort, int *validFrom
  );
  virtual bool statUser(const char *username, char *institution, int maxInstitutionLen, char *password, int maxPasswordLen);
};

class DynamicDatabase : public Database {
protected:
  struct planetStruct {
    char name[200];
    in_addr_t ip;
    char sip[20];
    int port;
    char sport[10];
  };

  struct satelliteStruct {
    char email[200];
    char identifier[200];
    char planetPassword[200];
    int planetIndex;
  };
    
  struct satelliteStruct *satellite;
  struct planetStruct *planet;
  char inputFile[200];
  char theIdentifier[200];
  char theOwner[200];
  int numSatellites;
  int numPlanets;
  int maxSatellites;
  int startTime;
  Selector *selector;
  Log *log;

public:
  DynamicDatabase(ParameterFile *params, Selector *selector, Log *log);

  virtual void getSatelliteAllocation(const char *email, const char *identifier, const int nowTS,
    bool *isAllocated, char *expOwner, char *expIdentifier, unsigned char *planetPwd,
    int *planetPwdLen, char *planetName, in_addr_t *planetIP, int *planetPort, 
    in_addr_t *lastSatelliteIP, bool *allowMobility, int *nextRefreshTS
  );
  virtual void recordSatelliteCredits(const char *email, const char *identifier, int fromTS,
    int onlineSec, in_addr_t ip, int port, const char *domain, int version, long long bytesSent,
    long long bytesReceived
  );
  virtual bool statExperiment(const char *owner, const char *identifier, int *startTime, int *endTime,
    bool *dynAllocPlanets, bool *dynAllocSats, bool *allowSatMigration, int *satsLeftToAllocate,
    char *description, int descriptionMaxLen
  );
  virtual bool statPlanet(const char *expOwner, const char *expIdentifier, in_addr_t ip, int port, 
    char *planetAddress
  );
  virtual void addPlanet(const char *expOwner, const char *expIdentifier, const char *address, 
    in_addr_t ip, int port
  );
  virtual int getNumSatellitesInExperiment(const char *expOwner, const char *expIdentifier);
  virtual bool getMatrixForExperiment(const char *owner, const char *identifier, unsigned char *matrix, int matrixSize);
  virtual int getNumPlanetsInExperiment(const char *expOwner, const char *expIdentifier);
  virtual int getExperiments(int fromPeriod, int toPeriod, char *buf, int bufsize, int maxEntries, char ** const owners, char ** const identifiers);
  virtual int getPlanets(const char *expOwner, const char *expIdentifier, char *buf, int bufsize, int maxEntries, char ** const names, char ** const ips, char ** const ports);
  virtual bool addSatellite(const char *expOwner, const char *expIdentifier, 
    const char *satEmail, const char *satIdentifier, const char *planetPassword, const char *virtualIP
  );
  virtual bool addBinding(const char *expOwner, const char *expIdentifier, 
    const char *satEmail, const char *satIdentifier, const char *satIP, 
    const char *planetAddress, int planetPort
  );
  virtual bool findBinding(const char *expOwner, const char *expIdentifier,
    const char *satEmail, const char *satIdentifier, in_addr_t satIP, 
    char *planetAddress, in_addr_t *planetIP, int *planetPort, int *validFrom
  );
  virtual bool statUser(const char *username, char *institution, int maxInstitutionLen, char *password, int maxPasswordLen);
};

#endif /* defined(__sun_h__) */
