#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <netdb.h>

#include "sun.h"
#include "tools.h"

DynamicDatabase::DynamicDatabase(ParameterFile *params, Selector *selector, Log *log) : Database()
{
  this->selector = selector;
  this->log = log;
  this->startTime = time(0);

  if (params->existsSetting("experiment"))
    params->getAsString("experiment", theIdentifier, sizeof(theIdentifier));
  else
    strcpy(theIdentifier, "default");

  if (params->existsSetting("owner"))
    params->getAsString("owner", theOwner, sizeof(theOwner));
  else
    strcpy(theOwner, "local");
    
  if (params->existsSetting("maxSatellites"))
    maxSatellites = params->getAsInt("maxSatellites");
  else
    maxSatellites = 50;

  satellite = (struct satelliteStruct*)NULL;
  numSatellites = 0;
  planet = (struct planetStruct*)NULL;
  numPlanets = 0;
}

void DynamicDatabase::getSatelliteAllocation(const char *email, const char *identifier, const int nowTS,
    bool *isAllocated, char *expOwner, char *expIdentifier, unsigned char *planetPwd,
    int *planetPwdLen, char *planetName, in_addr_t *planetIP, int *planetPort, 
    in_addr_t *lastSatelliteIP, bool *allowMobility, int *nextRefreshTS)
{
  *isAllocated = false;
  if (expOwner)
    expOwner[0] = 0;
  if (expIdentifier)
    expIdentifier[0] = 0;
  if (planetPwd)
    planetPwd[0] = 0;
  if (planetPwdLen)
    *planetPwdLen = 0;
  if (planetName)
    planetName[0] = 0;
  if (planetIP)
    *planetIP = 0;
  if (planetPort)
    *planetPort = 0;
  if (lastSatelliteIP)
    *lastSatelliteIP = 0;
  if (allowMobility)
    *allowMobility = false;
  if (nextRefreshTS)
    *nextRefreshTS = time(0)+3600;
  
  for (int i=0; i<numSatellites; i++) {
    if (!strcmp(satellite[i].email, email) && !strcmp(satellite[i].identifier, identifier)) {
      *isAllocated = true;
      if (expOwner)
        strcpy(expOwner, theOwner);
      if (expIdentifier)
        strcpy(expIdentifier, theIdentifier);
      if (planetPwd) {
        memset(planetPwd, 0, 8);
        planetPwd[0] = 0xAA;
        *(unsigned int*)&planetPwd[4] = htonl(1+i);
      }
      if (planetPwdLen)
        *planetPwdLen = 8;
      if (satellite[i].planetIndex >= 0) {
        if (planetName) 
          strcpy(planetName, planet[satellite[i].planetIndex].name);
        if (planetIP)
          *planetIP = planet[satellite[i].planetIndex].ip;
        if (planetPort)
          *planetPort = planet[satellite[i].planetIndex].port;
      } else {
        if (planetName)
          *planetName = 0;
        if (planetIP)
          *planetIP = 0;
        if (planetPort)
          *planetPort = 0;
      }
    }
  }
}

void DynamicDatabase::recordSatelliteCredits(const char *email, const char *identifier, int fromTS,
    int onlineSec, in_addr_t ip, int port, const char *domain, int version, long long bytesSent,
    long long bytesReceived)
{
  /* we don't do this in this case */
}

bool DynamicDatabase::statExperiment(const char *owner, const char *identifier, int *startTimeArg, int *endTime,
    bool *dynAllocPlanets, bool *dynAllocSats, bool *allowSatMigration, int *satsLeftToAllocate,
    char *description, int descriptionMaxLen)
{
  if (strcmp(owner, theOwner) || strcmp(identifier, theIdentifier))
    return false;
    
  if (startTimeArg)
    *startTimeArg = startTime;
  if (endTime)
    *endTime = startTime+100000*3600;
  if (dynAllocPlanets)
    *dynAllocPlanets = true;
  if (dynAllocSats)
    *dynAllocSats = true;
  if (allowSatMigration)
    *allowSatMigration = false;
  if (satsLeftToAllocate)
    *satsLeftToAllocate = maxSatellites - numSatellites;
  if (description) 
    snprintf(description, descriptionMaxLen, "Dynamic experiment", inputFile);
    
  return true;
}

bool DynamicDatabase::statPlanet(const char *expOwner, const char *expIdentifier, in_addr_t ip, int port, char *planetAddress)
{
  if (strcmp(expOwner, theOwner) || strcmp(expIdentifier, theIdentifier))
    return false;
    
  for (int i=0; i<numPlanets; i++) {
    if ((planet[i].ip == ip) && (planet[i].port == port)) {
      strcpy(planetAddress, planet[i].name);
      return true;
    }
  }
  
  return false;
}

void DynamicDatabase::addPlanet(const char *expOwner, const char *expIdentifier, const char *address, in_addr_t ip, int port)
{
  if (strcmp(expOwner, theOwner) || strcmp(expIdentifier, theIdentifier))
    panic("addPlanet: Dynamic database allows only a single experiment (%s:%s)", theOwner, theIdentifier);

  struct planetStruct *newStruct = (struct planetStruct*) malloc((numPlanets+1) * sizeof(struct planetStruct));
  if (numPlanets>0)
    memcpy(newStruct, planet, numPlanets*sizeof(struct planetStruct));
  if (planet)
    free(planet);
  
  planet = newStruct;
  int idx = numPlanets ++;
  strncpy(planet[idx].name, address, sizeof(planet[idx].name));
  planet[idx].ip = ip;
  ipdisp(ip, planet[idx].sip);
  planet[idx].port = port;
  sprintf(planet[idx].sport, "%d", port);
}

int DynamicDatabase::getNumSatellitesInExperiment(const char *expOwner, const char *expIdentifier)
{
  return numSatellites;
}

bool DynamicDatabase::getMatrixForExperiment(const char *owner, const char *identifier, unsigned char *matrix, int matrixSize)
{
log->write(1, "getMatrix %d", matrixSize);
  if (strcmp(owner, theOwner) || strcmp(identifier, theIdentifier))
    return false;

  int numRows = (matrixSize>numSatellites) ? numSatellites : matrixSize;
  for (int i=0; i<numRows; i++) {
    unsigned char *base = &matrix[18*i];
    memset(base, 0, 18);
    for (int j=0; j<8; j++)
      base[j] = 0;
    int ppl = strlen(satellite[i].planetPassword)/2;
    if (ppl>8)
      ppl = 8;
    for (int j=0; j<ppl; j++) {
      char s[3] = { satellite[i].planetPassword[2*j], satellite[i].planetPassword[2*j+1], 0 };
      base[j] = strtoul(s, 0, 16);
    }

    if (satellite[i].planetIndex >= 0) {
      *(in_addr_t*)&base[8] = planet[satellite[i].planetIndex].ip;
      *(unsigned short*)&base[12] = htons(planet[satellite[i].planetIndex].port);
    } else {
      *(in_addr_t*)&base[8] = 0;
      *(unsigned short*)&base[12] = 0;
    }
    *(in_addr_t*)&base[14] = 0;
  }
  
  return true;
}

int DynamicDatabase::getNumPlanetsInExperiment(const char *expOwner, const char *expIdentifier)
{
  return numPlanets;
}

int DynamicDatabase::getExperiments(int fromPeriod, int toPeriod, char *buf, int bufsize, int maxEntries, char ** const owners, char ** const identifiers)
{
  owners[0] = theOwner;
  identifiers[0] = theIdentifier;
  return 1;
}

int DynamicDatabase::getPlanets(const char *expOwner, const char *expIdentifier, char *buf, int bufsize, int maxEntries, char ** const names, char ** const ips, char ** const ports)
{
  if (strcmp(expOwner, theOwner) || strcmp(expIdentifier, theIdentifier))
    return -1;
    
  int numRows = (maxEntries > numPlanets) ? numPlanets : maxEntries;
  for (int i=0; i<numRows; i++) {
    names[i] = planet[i].name;
    ips[i] = planet[i].sip;
    ports[i] = planet[i].sport;
  }
  
  return numRows;;
}

bool DynamicDatabase::addSatellite(const char *expOwner, const char *expIdentifier, const char *satEmail, const char *satIdentifier, const char *planetPassword, const char *virtualIP)
{
log->write(1, "addsat %d", numSatellites+1);
  if (strcmp(expOwner, theOwner) || strcmp(expIdentifier, theIdentifier))
    panic("addSatellite: Dynamic database allows only a single experiment (%s:%s)", theOwner, theIdentifier);

  struct satelliteStruct *newStruct = (struct satelliteStruct*) malloc((numSatellites+1) * sizeof(struct satelliteStruct));
  if (numSatellites>0)
    memcpy(newStruct, satellite, numSatellites*sizeof(struct satelliteStruct));
  if (satellite)
    free(satellite);
  
  satellite = newStruct;
  int idx = numSatellites ++;
  strncpy(satellite[idx].email, satEmail, sizeof(satellite[idx].email));
  strncpy(satellite[idx].identifier, satIdentifier, sizeof(satellite[idx].identifier));
  strncpy(satellite[idx].planetPassword, planetPassword, sizeof(satellite[idx].planetPassword));
  satellite[idx].planetIndex = -1;
}

bool DynamicDatabase::addBinding(const char *expOwner, const char *expIdentifier, 
    const char *satEmail, const char *satIdentifier, const char *satIP, 
    const char *planetAddress, int planetPort)
{
  if (strcmp(expOwner, theOwner) || strcmp(expIdentifier, theIdentifier))
    return false;
  
  int planetIndex = -1;
  for (int i=0; i<numPlanets; i++) {
    if (!strcmp(planet[i].name, planetAddress) && (planet[i].port == planetPort))
      planetIndex = i;
  }
  
  if (planetIndex < 0)
    panic("addBinding: Trying to bind to an unregistered planet: %s:%d", planetAddress, planetPort);
  
  for (int i=0; i<numSatellites; i++) {
    if (!strcmp(satellite[i].email, satEmail) && !strcmp(satellite[i].identifier, satIdentifier)) {
      satellite[i].planetIndex = planetIndex;
      return true;
    }
  }
  
  return true;
}

bool DynamicDatabase::findBinding(const char *expOwner, const char *expIdentifier,
    const char *satEmail, const char *satIdentifier, in_addr_t satIP, 
    char *planetAddress, in_addr_t *planetIP, int *planetPort, int *validFrom)
{
  if (strcmp(expOwner, theOwner) || strcmp(expIdentifier, theIdentifier))
    return false;
    
  for (int i=0; i<numSatellites; i++) {
    if ((satellite[i].planetIndex >= 0) && !strcmp(satellite[i].email, satEmail) && !strcmp(satellite[i].identifier, satIdentifier)) {
      strcpy(planetAddress, planet[satellite[i].planetIndex].name);
      *planetIP = planet[satellite[i].planetIndex].ip;
      *planetPort = planet[satellite[i].planetIndex].port;
      *validFrom = startTime;
      return true;
    }
  }
  
  return false;
}

bool DynamicDatabase::statUser(const char *username, char *institution, int maxInstitutionLen, char *password, int maxPasswordLen)
{
  return false;
}
