#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <netdb.h>

#include "sun.h"
#include "tools.h"

FileDatabase::FileDatabase(ParameterFile *params, Selector *selector, Log *log) : Database()
{
  this->selector = selector;
  this->log = log;
  this->startTime = time(0);

  if (params->existsSetting("experiment"))
    params->getAsString("experiment", theIdentifier, sizeof(theIdentifier));
  else
    strcpy(theIdentifier, "default");

  if (params->existsSetting("owner"))
    params->getAsString("owner", theOwner, sizeof(theOwner));
  else
    strcpy(theOwner, "local");

  params->getAsString("databaseFile", inputFile, sizeof(inputFile));

  FILE *infile = fopen(inputFile, "r");
  if (!infile)
    panic("Cannot read allocation from '%s'", inputFile);
  
  log->write(1, "Reading allocation from '%s'", inputFile);
  
  satellite = (struct satelliteStruct*)malloc(0);
  numSatellites = 0;
  planet = (struct planetStruct*)malloc(0);
  numPlanets = 0;
  
  char linebuf[1000];
  int lineno = 0, satelliteno = 0;
  while (fgets(linebuf, sizeof(linebuf), infile)) {
    lineno ++;
    if (linebuf[0] == '#')
      continue;
      
    char *arg[4];
    arg[0] = strtok(linebuf, ",");
    arg[1] = arg[0] ? strtok(NULL, ",") : NULL;
    arg[2] = arg[1] ? strtok(NULL, ":") : NULL;
    arg[3] = arg[2] ? strtok(NULL, "\r\n") : NULL;
    
    if (!arg[3])
      panic("Cannot parse line %s:%d", inputFile, lineno);

    int planetPort = atoi(arg[3]);

    int pi = -1;
    for (int i=0; i<numPlanets; i++) {
      if (!strcmp(planet[i].name, arg[2]) && (planet[i].port == planetPort))
        pi = i;
    }
    
    if (pi < 0) {
      struct planetStruct *newPlanet = (struct planetStruct*)malloc((numPlanets+1)*sizeof(struct planetStruct));
      memcpy(newPlanet, planet, numPlanets * sizeof(struct planetStruct));
      free(planet);
      planet = newPlanet;
      pi = numPlanets ++;
      strncpy(planet[pi].name, arg[2], sizeof(planet[pi].name));
      planet[pi].port = planetPort;

      struct hostent *ent = gethostbyname(arg[2]);
      if (!ent)
        panic("Cannot resolve planet address '%s' at %s:%d", arg[2], inputFile, lineno);
        
      planet[pi].ip = *(in_addr_t*)ent->h_addr;
      ipdisp(planet[pi].ip, planet[pi].sip);
      sprintf(planet[pi].sport, "%d", planet[pi].port);
    }

    struct satelliteStruct *newSatellite = (struct satelliteStruct*)malloc((numSatellites+1)*sizeof(struct satelliteStruct));
    memcpy(newSatellite, satellite, numSatellites * sizeof(struct satelliteStruct));
    free(satellite);
    satellite = newSatellite;
    int idx = numSatellites ++;
    
    strcpy(satellite[idx].email, arg[0]);
    strcpy(satellite[idx].identifier, arg[1]);
    satellite[idx].planetIndex = pi;
  }
  
  fclose(infile);
}

void FileDatabase::getSatelliteAllocation(const char *email, const char *identifier, const int nowTS,
    bool *isAllocated, char *expOwner, char *expIdentifier, unsigned char *planetPwd,
    int *planetPwdLen, char *planetName, in_addr_t *planetIP, int *planetPort, 
    in_addr_t *lastSatelliteIP, bool *allowMobility, int *nextRefreshTS)
{
  *isAllocated = false;
  if (expOwner)
    expOwner[0] = 0;
  if (expIdentifier)
    expIdentifier[0] = 0;
  if (planetPwd)
    planetPwd[0] = 0;
  if (planetPwdLen)
    *planetPwdLen = 0;
  if (planetName)
    planetName[0] = 0;
  if (planetIP)
    *planetIP = 0;
  if (planetPort)
    *planetPort = 0;
  if (lastSatelliteIP)
    *lastSatelliteIP = 0;
  if (allowMobility)
    *allowMobility = false;
  if (nextRefreshTS)
    *nextRefreshTS = time(0)+3600;
  
  for (int i=0; i<numSatellites; i++) {
    if (!strcmp(satellite[i].email, email) && !strcmp(satellite[i].identifier, identifier)) {
      *isAllocated = true;
      if (expOwner)
        strcpy(expOwner, theOwner);
      if (expIdentifier)
        strcpy(expIdentifier, theIdentifier);
      if (planetPwd) {
        memset(planetPwd, 0, 8);
        planetPwd[0] = 0xAA;
        *(unsigned int*)&planetPwd[4] = htonl(1+i);
      }
      if (planetPwdLen)
        *planetPwdLen = 8;
      if (planetName)
        strcpy(planetName, planet[satellite[i].planetIndex].name);
      if (planetIP)
        *planetIP = planet[satellite[i].planetIndex].ip;
      if (planetPort)
        *planetPort = planet[satellite[i].planetIndex].port;
    }
  }
}

void FileDatabase::recordSatelliteCredits(const char *email, const char *identifier, int fromTS,
    int onlineSec, in_addr_t ip, int port, const char *domain, int version, long long bytesSent,
    long long bytesReceived)
{
  /* we don't do this in this case */
}

bool FileDatabase::statExperiment(const char *owner, const char *identifier, int *startTimeArg, int *endTime,
    bool *dynAllocPlanets, bool *dynAllocSats, bool *allowSatMigration, int *satsLeftToAllocate,
    char *description, int descriptionMaxLen)
{
  if (strcmp(owner, theOwner) || strcmp(identifier, theIdentifier))
    return false;
    
  if (startTimeArg)
    *startTimeArg = startTime;
  if (endTime)
    *endTime = startTime+100000*3600;
  if (dynAllocPlanets)
    *dynAllocPlanets = false;
  if (dynAllocSats)
    *dynAllocSats = false;
  if (allowSatMigration)
    *allowSatMigration = false;
  if (satsLeftToAllocate)
    *satsLeftToAllocate = 0;
  if (description) 
    snprintf(description, descriptionMaxLen, "Experiment loaded from '%s'", inputFile);
    
  return true;
}

bool FileDatabase::statPlanet(const char *expOwner, const char *expIdentifier, in_addr_t ip, int port, char *planetAddress)
{
  if (strcmp(expOwner, theOwner) || strcmp(expIdentifier, theIdentifier))
    return false;
    
  for (int i=0; i<numPlanets; i++) {
    if ((planet[i].ip == ip) && (planet[i].port == port)) {
      strcpy(planetAddress, planet[i].name);
      return true;
    }
  }
  
  return false;
}

void FileDatabase::addPlanet(const char *expOwner, const char *expIdentifier, const char *address, in_addr_t ip, int port)
{
  panic("Cannot add planets when the static database is used");
}

int FileDatabase::getNumSatellitesInExperiment(const char *expOwner, const char *expIdentifier)
{
  return numSatellites;
}

bool FileDatabase::getMatrixForExperiment(const char *owner, const char *identifier, unsigned char *matrix, int matrixSize)
{
  if (strcmp(owner, theOwner) || strcmp(identifier, theIdentifier))
    return false;

  int numRows = (matrixSize>numSatellites) ? numSatellites : matrixSize;
  for (int i=0; i<numRows; i++) {
    unsigned char *base = &matrix[18*i];
    memset(base, 0, 18);
    base[0] = 0xAA;
    *(unsigned int*)&base[4] = htonl(1+i);
    *(in_addr_t*)&base[8] = planet[satellite[i].planetIndex].ip;
    *(unsigned short*)&base[12] = htons(planet[satellite[i].planetIndex].port);
    *(in_addr_t*)&base[14] = 0;
  }
  
  return true;
}

int FileDatabase::getNumPlanetsInExperiment(const char *expOwner, const char *expIdentifier)
{
  return numPlanets;
}

int FileDatabase::getExperiments(int fromPeriod, int toPeriod, char *buf, int bufsize, int maxEntries, char ** const owners, char ** const identifiers)
{
  owners[0] = theOwner;
  identifiers[0] = theIdentifier;
  return 1;
}

int FileDatabase::getPlanets(const char *expOwner, const char *expIdentifier, char *buf, int bufsize, int maxEntries, char ** const names, char ** const ips, char ** const ports)
{
  if (strcmp(expOwner, theOwner) || strcmp(expIdentifier, theIdentifier))
    return -1;
    
  int numRows = (maxEntries > numPlanets) ? numPlanets : maxEntries;
  for (int i=0; i<numRows; i++) {
    names[i] = planet[i].name;
    ips[i] = planet[i].sip;
    ports[i] = planet[i].sport;
  }
  
  return numRows;;
}

bool FileDatabase::addSatellite(const char *expOwner, const char *expIdentifier, const char *satEmail, const char *satIdentifier, const char *planetPassword, const char *virtualIP)
{
  panic("Cannot add satellites when the static database is used");
}

bool FileDatabase::addBinding(const char *expOwner, const char *expIdentifier, 
    const char *satEmail, const char *satIdentifier, const char *satIP, 
    const char *planetAddress, int planetPort)
{
  panic("Cannot add bindings when the static database is used");
}

bool FileDatabase::findBinding(const char *expOwner, const char *expIdentifier,
    const char *satEmail, const char *satIdentifier, in_addr_t satIP, 
    char *planetAddress, in_addr_t *planetIP, int *planetPort, int *validFrom)
{
  if (strcmp(expOwner, theOwner) || strcmp(expIdentifier, theIdentifier))
    return false;
    
  for (int i=0; i<numSatellites; i++) {
    if (!strcmp(satellite[i].email, satEmail) && !strcmp(satellite[i].identifier, satIdentifier)) {
      strcpy(planetAddress, planet[satellite[i].planetIndex].name);
      *planetIP = planet[satellite[i].planetIndex].ip;
      *planetPort = planet[satellite[i].planetIndex].port;
      *validFrom = startTime;
      return true;
    }
  }
  
  return false;
}

bool FileDatabase::statUser(const char *username, char *institution, int maxInstitutionLen, char *password, int maxPasswordLen)
{
  return false;
}
