#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <sys/wait.h>
#include <netdb.h>
#include <errno.h>

#include "sun.h"
#include "tools.h"

int MysqlDatabase::mysqlQuery(const char *dbName, const char *username, const char *passwd, const char *query, char *buf, int maxlen, char **arg, int maxColumns, int maxEntries)
{
  log->write(4, "Query: [%s]", query);

  int outpipe[2], errpipe[2];
  pipe(outpipe);
  pipe(errpipe);
  
  /* Fork a process that will execute the 'mysql' programm with the specified query.
     We pipe the output of this process to a pipe, where it can be read by the
     main collector process. */

  pid_t child = fork();
  if (child == 0) {
    close(outpipe[0]);
    close(errpipe[0]);
    if (outpipe[1] != STDOUT_FILENO) {
      close(STDOUT_FILENO);
      if (dup2(outpipe[1], STDOUT_FILENO) < 0)
        panic("Cannot dup");
    }
    if (errpipe[1] != STDERR_FILENO) {
      close(STDERR_FILENO);
      if (dup2(errpipe[1], STDERR_FILENO) < 0)
        panic("Cannot dup");
    }
    
    char userarg[200], passwdarg[200];
    snprintf(userarg, sizeof(userarg), "-u%s", username);
    snprintf(passwdarg, sizeof(passwdarg), "-p%s", passwd);
    execlp("mysql", "mysql", dbName, userarg, passwdarg, "-e", query, NULL);
    exit(1);
  } 
  
  close(outpipe[1]);
  close(errpipe[1]);
  
  /* Read until the end of file is reached (i.e. the mysql process has terminated) */
  
  int p = 0;
  while (p<(maxlen-1)) {
    int r = read(outpipe[0], buf, (maxlen-1)-p);
    if (r<0) {
      log->write(1, "ERROR: Failed to read from mysql child process (%s)", strerror(errno));
      panic("Cannot read from mysql child process");
    }
    if (r == 0)
      break;

    p += r;
  }
  
  char err[200];
  int r = read(errpipe[0], err, sizeof(err)-1);
  if (r>0) {
    char *x;
    err[r] = 0;
    strtok_r(err, "\r\n", &x);
  } else {
    err[0] = 0;
  }
  
  close(outpipe[0]);
  close(errpipe[0]);
  
  assert(p<maxlen);
  
  buf[p] = 0;
  log->write(4, "Result: [%s]", buf);

  /* Clean up the child process descriptor */

  int status;
  wait(&status);
  if (status != 0)
    log->write(3, "WARNING: MySQL query '%s' appears to have failed (status %d, [%s])", query, status, err);
  
  /* Parse result */
  
  if (strlen(buf) < 1)
    return 0;
  
  char *p1, *p2;
  strtok_r(buf, "\r\n", &p1); // gets rid of the title row

  char *c = strtok_r(NULL, "\r\n", &p1);
  int argptr = 0, numRows = 0;
  while (c) {
    char *item[maxColumns];
    item[0] = strtok_r(c, "\t", &p2);
    for (int i=1; i<maxColumns; i++)
      item[i] = item[i-1] ? strtok_r(NULL, "\t", &p2) : NULL;
    
    for (int i=0; i<maxColumns; i++) {
      if (argptr < maxEntries) 
        arg[argptr++] = item[i];
    }

    c = strtok_r(NULL, "\r\n", &p1);
    numRows ++;
  }
  
  return numRows;
}

MysqlDatabase::MysqlDatabase(ParameterFile *params, Selector *selector, Log *log) : Database()
{
  this->selector = selector;
  this->log = log;

  params->getAsString("mysqlDbName", dbName, sizeof(dbName));
  params->getAsString("mysqlUsername", dbUsername, sizeof(dbUsername));
  params->getAsString("mysqlPassword", dbPassword, sizeof(dbPassword));
}

void MysqlDatabase::getSatelliteAllocation(const char *email, const char *identifier, const int nowTS, bool *isAllocated, char *expOwner, char *expIdentifier, unsigned char *planetPwd, int *planetPwdLen, char *planetName, in_addr_t *planetIP, int *planetPort, in_addr_t *lastSatelliteIP, bool *allowMigration, int *nextRefreshTS)
{
  char query[10000], result[10000];
  int now = time(0);
  
  snprintf(query, sizeof(query), 
    "SELECT experiments.owner, experiments.identifier, satellites.planetPassword, bindings.planetAddress, planets.ip, bindings.planetPort, experiments.endTime, bindings.satIP, experiments.allowSatMigration FROM satellites "
    "LEFT JOIN experiments ON satellites.experiment=experiments.identifier AND satellites.owner=experiments.owner "
    "LEFT JOIN bindings ON bindings.experiment=experiments.identifier AND bindings.owner=experiments.owner AND bindings.satEmail=satellites.email AND bindings.satIdentifier=satellites.identifier "
    "LEFT JOIN planets ON bindings.owner=planets.owner AND bindings.experiment=planets.experiment AND bindings.planetAddress=planets.address AND bindings.planetPort=planets.port "
    "WHERE satellites.email='%s' AND satellites.identifier='%s' AND endTime>%d "
    "ORDER BY startTime ASC, bindings.validFrom DESC LIMIT 1;",
    email, identifier, nowTS
  );
  
  char *arg[50];
  int numRows = mysqlQuery(dbName, dbUsername, dbPassword, query, result, sizeof(result), arg, 9, 50);

  *isAllocated = false;
  *expOwner = 0;
  *expIdentifier = 0;
  *planetPwdLen = 0;
  *planetName = 0;
  *planetIP = 0;
  *planetPort = 0;
  *lastSatelliteIP = 0;
  *allowMigration = false;
  *nextRefreshTS = now + 3600;
  
  if (numRows < 1)
    return;
    
  assert((numRows == 1) && arg[8]);
  
  *isAllocated = true;
  strncpy(expOwner, arg[0], 200);
  strncpy(expIdentifier, arg[1], 200);
  
  *planetPwdLen = (strlen(arg[2])>16) ? 8 : (strlen(arg[2])/2);
  for (int i=0; i<*planetPwdLen; i++) {
    char s[3] = { arg[2][2*i], arg[2][2*i+1], 0 };
    planetPwd[i] = strtoul(s, NULL, 16);
  }

  *nextRefreshTS = atoi(arg[6]);
  
  if (strcmp(arg[3], "NULL")) {
    strncpy(planetName, arg[3], 200);
    *planetIP = inet_addr(arg[4]);
    *planetPort = atoi(arg[5]);
  }
  
  if (strcmp(arg[7], "NULL"))
    *lastSatelliteIP = inet_addr(arg[7]);
  else
    *lastSatelliteIP = 0;
    
  *allowMigration = (atoi(arg[8])!=0);
}

bool MysqlDatabase::statExperiment(const char *owner, const char *identifier, int *startTime, int *endTime, bool *dynAllocPlanets, bool *dynAllocSats, bool *allowSatMigration, int *satsLeftToAllocate, char *description, int descriptionMaxLen)
{
  char query[10000], result[10000];
  snprintf(query, sizeof(query), 
    "SELECT startTime, endTime, dynAllocPlanets, dynAllocSats, allowSatMigration, satsLeftToAllocate, description FROM experiments WHERE owner='%s' AND identifier='%s';",
    owner, identifier
  );
  
  char *arg[50];
  int numRows = mysqlQuery(dbName, dbUsername, dbPassword, query, result, sizeof(result), arg, 7, 50);
  if (numRows < 1)
    return false;
  
  assert(arg[5]);
    
  if (startTime)
    *startTime = atoi(arg[0]);
  if (endTime)
    *endTime = atoi(arg[1]);
  if (dynAllocPlanets)
    *dynAllocPlanets = (atoi(arg[2])!=0);
  if (dynAllocSats)
    *dynAllocSats = (atoi(arg[3])!=0);
  if (allowSatMigration)
    *allowSatMigration = (atoi(arg[4])!=0);
  if (satsLeftToAllocate)
    *satsLeftToAllocate = atoi(arg[5]);
  if (description) {
    if (arg[6])
      strncpy(description, arg[6], descriptionMaxLen);
    else
      strcpy(description, "");
  }

  return true;
}

bool MysqlDatabase::statPlanet(const char *expOwner, const char *expIdentifier, in_addr_t ip, int port, char *planetAddress)
{
  char query[10000], result[10000], buf1[256];
  snprintf(query, sizeof(query), 
    "SELECT address FROM planets WHERE owner='%s' AND experiment='%s' AND ip='%s' and port=%d;",
    expOwner, expIdentifier, ipdisp(ip, buf1), port
  );
  
  char *arg[50];
  int numRows = mysqlQuery(dbName, dbUsername, dbPassword, query, result, sizeof(result), arg, 1, 50);
  if (numRows < 1)
    return false;
  
  assert(arg[0]);
    
  if (planetAddress)
    strncpy(planetAddress, arg[0], 200);

  return true;
}

bool MysqlDatabase::findBinding(const char *expOwner, const char *expIdentifier, const char *satEmail, const char *satIdentifier, in_addr_t satIP, char *planetAddress, in_addr_t *planetIP, int *planetPort, int *validFrom)
{
  char query[10000], result[10000], buf1[256];
  snprintf(query, sizeof(query), 
    "SELECT bindings.planetAddress,planets.ip,bindings.planetPort,bindings.validFrom "
    "FROM bindings LEFT JOIN planets ON bindings.owner=planets.owner AND bindings.experiment=planets.experiment AND bindings.planetAddress=planets.address AND bindings.planetPort=planets.port "
    "WHERE bindings.owner='%s' AND bindings.experiment='%s' AND bindings.satEmail='%s' AND bindings.satIdentifier='%s' AND bindings.satIP='%s' "
    "ORDER BY bindings.validFrom DESC LIMIT 1;",
    expOwner, expIdentifier, satEmail, satIdentifier, ipdisp(satIP, buf1)
  );
  
  char *arg[50];
  int numRows = mysqlQuery(dbName, dbUsername, dbPassword, query, result, sizeof(result), arg, 4, 50);
  if (numRows < 1)
    return false;
    
  assert(arg[3]);
  strcpy(planetAddress, arg[0]);
  *planetIP = inet_addr(arg[1]);
  *planetPort = atoi(arg[2]);
  *validFrom = atoi(arg[3]);
  return true;
}

void MysqlDatabase::addPlanet(const char *expOwner, const char *expIdentifier, const char *address, in_addr_t ip, int port)
{
  char query[10000], result[10000], buf1[256];
  snprintf(query, sizeof(query), 
    "REPLACE INTO planets VALUES ('%s','%s','%s','%s',%d);",
    expOwner, expIdentifier, address, ipdisp(ip, buf1), port
  );
  
  char *arg[50];
  mysqlQuery(dbName, dbUsername, dbPassword, query, result, sizeof(result), arg, 1, 50);
}

int MysqlDatabase::getNumSatellitesInExperiment(const char *owner, const char *identifier)
{
  char query[10000], result[10000];
  snprintf(query, sizeof(query), 
    "SELECT count(identifier) FROM satellites WHERE owner='%s' AND experiment='%s';",
    owner, identifier
  );
  
  char *arg[10];
  int numRows = mysqlQuery(dbName, dbUsername, dbPassword, query, result, sizeof(result), arg, 1, 10);
  if (numRows < 1)
    return -1;

  return atoi(arg[0]);  
}

int MysqlDatabase::getNumPlanetsInExperiment(const char *owner, const char *identifier)
{
  char query[10000], result[10000];
  snprintf(query, sizeof(query), 
    "SELECT count(address) FROM planets WHERE owner='%s' AND experiment='%s';",
    owner, identifier
  );
  
  char *arg[10];
  int numRows = mysqlQuery(dbName, dbUsername, dbPassword, query, result, sizeof(result), arg, 1, 10);
  if (numRows < 1)
    return -1;

  return atoi(arg[0]);  
}

int MysqlDatabase::getExperiments(int fromPeriod, int toPeriod, char *buf, int bufsize, int maxEntries, char **owners, char **identifiers)
{
  char query[10000];
  snprintf(query, sizeof(query), 
    "SELECT owner, identifier FROM experiments WHERE endTime>%d AND startTime<%d;",
    fromPeriod, toPeriod
  );
  
  char *arg[maxEntries*2+1];
  int numRows = mysqlQuery(dbName, dbUsername, dbPassword, query, buf, bufsize, arg, 2, maxEntries*2);
  int rowsReturned = (numRows>maxEntries) ? maxEntries : numRows;
  for (int i=0; i<rowsReturned; i++) {
    owners[i] = arg[2*i];
    identifiers[i] = arg[2*i+1];
  }
  
  return rowsReturned;  
}

bool MysqlDatabase::getMatrixForExperiment(const char *owner, const char *identifier, unsigned char *matrix, int maxEntries)
{
  char query[10000], result[10000];
  int now = time(0);
  
  snprintf(query, sizeof(query), 
    "SELECT satellites.planetPassword,planets.ip,bindings.planetPort,satellites.virtualIP FROM satellites "
    "LEFT JOIN (SELECT satEmail,satIdentifier,experiment,owner,max(validFrom) AS mvf FROM bindings WHERE owner='%s' AND experiment='%s' GROUP BY satEmail,satIdentifier) AS xxx "
    "ON satellites.identifier=xxx.satIdentifier AND satellites.email=xxx.satEmail AND satellites.experiment=xxx.experiment AND satellites.owner=xxx.owner "
    "LEFT JOIN bindings ON satellites.identifier=bindings.satIdentifier AND satellites.email=bindings.satEmail AND satellites.owner=bindings.owner AND satellites.experiment=bindings.experiment AND xxx.mvf=bindings.validFrom "
    "LEFT JOIN planets ON bindings.owner=planets.owner AND bindings.experiment=planets.experiment AND bindings.planetAddress=planets.address AND bindings.planetPort=planets.port "
    "WHERE satellites.owner='%s' AND satellites.experiment='%s' "
    "ORDER BY rank ASC,planetPassword ASC;",
    owner, identifier, owner, identifier
  );
  
  char *arg[maxEntries*4+1];
  int numRows = mysqlQuery(dbName, dbUsername, dbPassword, query, result, sizeof(result), arg, 4, maxEntries*4);
  if (numRows >  maxEntries)
    numRows = maxEntries;
    
  for (int i=0; i<numRows; i++) {
    unsigned char *base = &matrix[18*i];
    memset(&base[0], 0, 8);
    int ppl = strlen(arg[4*i+0])/2;
    if (ppl > 8)
      ppl = 8;
    for (int j=0; j<ppl; j++) {
      char s[3] = { arg[4*i+0][2*j], arg[4*i+0][2*j+1], 0 };
      base[j] = strtoul(s, 0, 16);
    }
    *(in_addr_t*)&base[8] = (strcmp(arg[4*i+1], "NULL")) ? inet_addr(arg[4*i+1]) : 0;
    *(unsigned short*)&base[12] = (strcmp(arg[4*i+2], "NULL")) ? htons(atoi(arg[4*i+2])) : 0;
    *(in_addr_t*)&base[14] = (strcmp(arg[4*i+3], "NULL")) ? inet_addr(arg[4*i+3]) : 0;
  }
  
  return true;
}

void MysqlDatabase::recordSatelliteCredits(const char *email, const char *identifier, int fromTS, int onlineSec, in_addr_t ip, int port, const char *domain, int version, long long bytesSent, long long bytesReceived)
{
  char query[10000], result[10000], buf1[200];
  
  snprintf(query, sizeof(query), 
    "INSERT INTO credit VALUES ('%s','%s',%d,%d,'%s',%d,'%s',%d,%lld,%lld);",
    email, identifier, fromTS, onlineSec, ipdisp(ip, buf1), port, domain, version, bytesSent, bytesReceived
  );
  
  char *arg[10];
  mysqlQuery(dbName, dbUsername, dbPassword, query, result, sizeof(result), arg, 1, 10);
}

bool MysqlDatabase::addSatellite(const char *expOwner, const char *expIdentifier, const char *satEmail, const char *satIdentifier, const char *planetPassword, const char *virtualIP)
{
  char query[10000], result[10000], buf1[256];
  sprintf(query, "SELECT max(rank) FROM satellites WHERE owner='%s' AND experiment='%s';", expOwner, expIdentifier);

  char *arg[50];
  int nr = mysqlQuery(dbName, dbUsername, dbPassword, query, result, sizeof(result), arg, 1, 50);
  if (nr < 1)
    return false;
  
  int rank = atoi(arg[0])+1;
  
  snprintf(query, sizeof(query), 
    "REPLACE INTO satellites VALUES ('%s','%s','%s','%s','%s',%s%s%s,%d);",
    expOwner, expIdentifier, satEmail, satIdentifier, planetPassword, virtualIP ? "'" : "", virtualIP ? virtualIP : "null", virtualIP ? "'" : "", rank
  );
  
  mysqlQuery(dbName, dbUsername, dbPassword, query, result, sizeof(result), arg, 1, 50);

  snprintf(query, sizeof(query), 
    "UPDATE experiments SET satsLeftToAllocate=satsLeftToAllocate-1 WHERE owner='%s' AND identifier='%s' AND satsLeftToAllocate>0 AND dynAllocSats>0;",
    expOwner, expIdentifier
  );
  
  mysqlQuery(dbName, dbUsername, dbPassword, query, result, sizeof(result), arg, 1, 50);
  return true;
}

int MysqlDatabase::getPlanets(const char *expOwner, const char *expIdentifier, char *buf, int bufsize, int maxEntries, char **names, char **ips, char **ports)
{
  char query[10000];
  snprintf(query, sizeof(query), 
    "SELECT address, ip, port FROM planets WHERE owner='%s' AND experiment='%s';",
    expOwner, expIdentifier
  );
  
  char *arg[maxEntries*3+1];
  int numRows = mysqlQuery(dbName, dbUsername, dbPassword, query, buf, bufsize, arg, 3, maxEntries*3);
  int rowsReturned = (numRows>maxEntries) ? maxEntries : numRows;
  for (int i=0; i<rowsReturned; i++) {
    names[i] = arg[3*i];
    ips[i] = arg[3*i+1];
    ports[i] = arg[3*i+2];
  }
  
  return rowsReturned;  
}

bool MysqlDatabase::addBinding(const char *expOwner, const char *expIdentifier, const char *satEmail, const char *satIdentifier, const char *satIP, const char *planetAddress, int planetPort)
{
  char query[10000], result[10000];
  
  snprintf(query, sizeof(query), 
    "INSERT INTO bindings VALUES ('%s','%s','%s','%s','%s','%s',%d,%d);",
    expOwner, expIdentifier, satEmail, satIdentifier, satIP, planetAddress, planetPort, time(0)
  );
  
  char *arg[10];
  mysqlQuery(dbName, dbUsername, dbPassword, query, result, sizeof(result), arg, 1, 10);
  return true;
}

bool MysqlDatabase::statUser(const char *username, char *institution, int maxInstitutionLen, char *password, int maxPasswordLen)
{
  char query[10000], result[10000];
  snprintf(query, sizeof(query), 
    "SELECT institution, password FROM users WHERE username='%s';", username
  );
  
  char *arg[50];
  int numRows = mysqlQuery(dbName, dbUsername, dbPassword, query, result, sizeof(result), arg, 2, 50);
  if (numRows < 1)
    return false;
  
  assert(arg[1]);
    
  if (institution)
    strncpy(institution, arg[0], maxInstitutionLen);
  if (password)
    strncpy(password, arg[1], maxPasswordLen);

  return true;
}
