#include <sys/time.h>
#include <stdlib.h>
#include <errno.h>

#include "sun.h"

Selector::Selector()
{
  numDescriptors = 0;
  numTimers = 0;
}

long long Selector::getTimeMicros()
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec * 1000000LL + tv.tv_usec;
}

void Selector::setDescriptorCallback(int fd, bool wantRead, bool wantWrite, bool wantExceptions, SelectorCallback *callback)
{
  for (int i=0; i<numDescriptors; i++) {
    if ((descriptors[i].fd == fd) && (descriptors[i].callback == callback)) {
      descriptors[i].wantRead = wantRead;
      descriptors[i].wantWrite = wantWrite;
      descriptors[i].wantExc = wantExceptions;
      return;
    }
  }
  
  if (numDescriptors >= MAX_DESCRIPTORS)
    panic("Too many active descriptors");
    
  int idx = numDescriptors ++;
  descriptors[idx].fd = fd;
  descriptors[idx].wantRead = wantRead;
  descriptors[idx].wantWrite = wantWrite;
  descriptors[idx].wantExc = wantExceptions;
  descriptors[idx].callback = callback;
}

void Selector::cancelDescriptorCallback(int fd, SelectorCallback *callback)
{
  for (int i=0; i<numDescriptors; i++) {
    if ((descriptors[i].fd == fd) && (descriptors[i].callback == callback)) {
      descriptors[i] = descriptors[--numDescriptors];
      return;
    }
  }
}

void Selector::setTimer(int id, int arg, long long whenMicros, SelectorCallback *callback)
{
  for (int i=0; i<numTimers; i++) {
    if ((timers[i].id == id) && (timers[i].arg == arg) && (timers[i].callback == callback)) {
      timers[i].whenMicros = whenMicros;
      return;
    }
  }
  
  if (numTimers >= MAX_TIMERS)
    panic("Too many active timers");
    
  int idx = numTimers ++;
  timers[idx].id = id;
  timers[idx].arg = arg;
  timers[idx].whenMicros = whenMicros;
  timers[idx].callback = callback;
}

void Selector::run()
{
  long long now = getTimeMicros();
  while (true) {
    fd_set fdr, fdw, fdx;
    FD_ZERO(&fdr);
    FD_ZERO(&fdw);
    FD_ZERO(&fdx);
    
    int maxfd = 0;
    for (int i=0; i<numDescriptors; i++) {
      if (descriptors[i].wantRead)
        FD_SET(descriptors[i].fd, &fdr);
      if (descriptors[i].wantWrite)
        FD_SET(descriptors[i].fd, &fdw);
      if (descriptors[i].wantExc)
        FD_SET(descriptors[i].fd, &fdx);
      if (descriptors[i].fd > maxfd)
        maxfd = descriptors[i].fd;
    }

    long long earliest = -1;
    for (int i=0; i<numTimers; i++) {
      if ((earliest<0) || (timers[i].whenMicros < earliest))
        earliest = timers[i].whenMicros;
    }
    
    struct timeval tv;
    struct timeval *ptv = &tv;
    if (earliest < 0) {
      ptv = NULL;
    } else if (earliest<=now) {
      tv.tv_sec = 0;
      tv.tv_usec = 0;
    } else {
      long long toMicros = earliest-now;
      tv.tv_sec = toMicros / 1000000LL;
      tv.tv_usec = toMicros % 1000000LL;
    }
    
    int ret = select(maxfd + 1, &fdr, &fdw, &fdx, ptv);
    if (ret < 0) {
      if ((errno == EAGAIN) || (errno == ECONNREFUSED))
        continue;
        
      panic("Error on select; errno=%d", errno);
    }
    
    now = getTimeMicros();
    
    struct descriptor pendingDesc[numDescriptors];
    int numPendingDesc = 0;
      
    for (int i=0; i<numDescriptors; i++) {
      bool canRead = (descriptors[i].wantRead && FD_ISSET(descriptors[i].fd, &fdr));
      bool canWrite = (descriptors[i].wantWrite && FD_ISSET(descriptors[i].fd, &fdw));
      bool hasExc = (descriptors[i].wantExc && FD_ISSET(descriptors[i].fd, &fdx));
      if (canRead || canWrite || hasExc) {
        int idx = numPendingDesc ++;
        pendingDesc[idx].fd = descriptors[i].fd;
        pendingDesc[idx].wantRead = canRead;
        pendingDesc[idx].wantWrite = canWrite;
        pendingDesc[idx].wantExc = hasExc;
        pendingDesc[idx].callback = descriptors[i].callback;
      }
    }
    
    struct timer pendingTimers[numTimers];
    int numPendingTimers = 0;
    
    for (int i=0; i<numTimers; ) {
      if (timers[i].whenMicros <= now) {
        int idx = numPendingTimers ++;
        pendingTimers[idx] = timers[i];
        timers[i] = timers[--numTimers];
      } else {
        i++;
      }
    }
    
    for (int i=0; i<numPendingDesc; i++)
      pendingDesc[i].callback->descriptorReady(pendingDesc[i].fd, pendingDesc[i].wantRead, pendingDesc[i].wantWrite, pendingDesc[i].wantExc);
    for (int i=0; i<numPendingTimers; i++)
      pendingTimers[i].callback->timerExpired(pendingTimers[i].id, pendingTimers[i].arg);      
  }
}
