#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <time.h>
#include <ctype.h>
#include <netdb.h>
#include <assert.h>
#include <openssl/sha.h>

#include "sun.h"
#include "tools.h"
#include "parameters.h"

Sun::Sun(ParameterFile *params, Selector *selector, Log *log, Database *db) : SelectorCallback()
{
  this->selector = selector;
  this->log = log;
  this->db = db;
  this->masterSocket = -1;
  this->masterPort = params->existsSetting("port") ? params->getAsInt("port") : DEFAULT_MASTER_PORT;
  this->numSatellites = 0;
  this->numPlanets = 0;
  this->numExperiments = 0;
  this->numMatrixCacheEntries = 0;
  this->nextBootstrapID = random();
}

void Sun::init()
{
  masterSocket = openServerSocket(masterPort, SOCK_DGRAM);
  if (masterSocket<0)
    panic("Cannot open server socket %d/udp (%s)", masterPort, strerror(errno));
  
  selector->setDescriptorCallback(masterSocket, true, false, false, this);
  selector->setTimer(TI_HOUSEKEEPING, 0, selector->getTimeMicros()+HOUSEKEEPING_INTERVAL_MICROS, this);
  selector->setTimer(TI_FLUSH_CREDIT, 0, selector->getTimeMicros()+FLUSH_CREDIT_INTERVAL_MICROS, this);
  selector->setTimer(TI_EXPERIMENT_REFRESH, 0, selector->getTimeMicros()+1, this);
  
  log->write(1, "Starting %s %s", PACKAGE, VERSION);
  log->write(2, "Server socket is %d/udp", masterPort);
}

int Sun::getBestNextBootstrap(int idx)
{
  int best = -1;
  for (int i=0; i<satellite[idx].numBootstraps; i++) {
    if (satellite[idx].bootAttempts[i] < NUM_BOOTSTRAP_ATTEMPTS_PER_NODE)
      if ((best<0) || (satellite[idx].bootAttempts[i]<satellite[idx].bootAttempts[best]))
        best = i;
  }
  
  return best;
}

void Sun::finishBootstrappingSatellite(int idx)
{
  assert(satellite[idx].isBootstrapping && satellite[idx].isAllocated);

  /* Find the planet with the lowest RTT to the satellite */

  int bestIdx = -1;
  int bestRTT = 9999999;
  for (int i=0; i<satellite[idx].numBootstraps; i++) {
    log->write(3, "Bootstrap %s:%d: %d/%d successful attempts, minRTT=%dms", 
      satellite[idx].bootAddress[i], satellite[idx].bootPort[i], satellite[idx].bootSuccesses[i], satellite[idx].bootAttempts[i], 
      satellite[idx].bootBestRTT[i]
    );
    
    if (satellite[idx].bootSuccesses[i] > 0) {
      if ((bestIdx<0) || (bestRTT > satellite[idx].bootBestRTT[i])) {
        bestIdx = i;
        bestRTT = satellite[idx].bootBestRTT[i];
      }
    }
  }
  
  if (bestIdx >= 0) {
  
    /* Typical case: Bootstrap was successful; associate the satellite with the planet */
  
    satellite[idx].planetIP = satellite[idx].bootIP[bestIdx];
    satellite[idx].planetPort = satellite[idx].bootPort[bestIdx];
    strncpy(satellite[idx].planetName, satellite[idx].bootAddress[bestIdx], sizeof(satellite[idx].planetName));
    
    log->write(1, "Satellite '%s:%s' is now associated with planet %s:%d (RTT %dms)", 
      satellite[idx].email, satellite[idx].identifier, satellite[idx].planetName, satellite[idx].planetPort, bestRTT
    );
    
    char buf1[20];
    db->addBinding(satellite[idx].expOwner, satellite[idx].expIdentifier, satellite[idx].email, satellite[idx].identifier, 
      ipdisp(satellite[idx].ip, buf1), satellite[idx].planetName, satellite[idx].planetPort
    );
    
    /* We've just changed the matrix of this experiment, so invalidate any cache entries */
    
    invalidateMatrix(satellite[idx].expOwner, satellite[idx].expIdentifier);
  } else {
  
    /* If the satellite can't talk to any planets at all, disable it */
  
    log->write(1, "Satellite '%s:%s' can't talk to any bootstrap node in experiment '%s:%s' -- disabling", 
      satellite[idx].email, satellite[idx].identifier, satellite[idx].expOwner, satellite[idx].expIdentifier
    );
    
    satellite[idx].planetIP = 0;
    satellite[idx].planetPort = 0;
    satellite[idx].planetName[0] = 0;
  }

  cancelBootstrappingSatellite(idx);
}

void Sun::getGoalSettingsForSatellite(int idx, int *sunInterval, int *planetInterval, in_addr_t *planetIp, int *planetPort, unsigned char *pwdBuf, unsigned int pwdMaxLen, int *pwdLen)
{
  assert((0<=idx) && (idx<numSatellites));
  
  /* During normal operation, the satellite reports to the planet every second, and to the sun every 60 seconds */
  
  if (!satellite[idx].isBootstrapping) {
    *sunInterval = 60;
    *planetInterval = 1;
    if (satellite[idx].isAllocated) {
      *planetIp = satellite[idx].planetIP;
      *planetPort = satellite[idx].planetPort;
    } else {
      *planetIp = 0;
      *planetPort = 0;
    } 
    
    *pwdLen = (pwdMaxLen>satellite[idx].planetPasswordLen) ? satellite[idx].planetPasswordLen : pwdMaxLen;
    memcpy(pwdBuf, satellite[idx].planetPassword, *pwdLen);
  } else {
  
    /* While bootstrapping, the satellite reports to the sun once per second. During the bootstrap phase, the
       sun sends new commands to the satellite every 1-2 seconds, so we need this to keep any NATs or firewalls open.  */
  
    assert((0<=satellite[idx].currentBootstrapIndex) && (satellite[idx].currentBootstrapIndex<satellite[idx].numBootstraps));
    *sunInterval = 1;
    *planetInterval = 2;
    *planetIp = satellite[idx].bootIP[satellite[idx].currentBootstrapIndex];
    *planetPort = satellite[idx].bootPort[satellite[idx].currentBootstrapIndex];
    
    assert(pwdMaxLen>=5);
    *pwdLen = 5;
    pwdBuf[0] = 0xBB;
    *(unsigned int*)&pwdBuf[1] = htonl(satellite[idx].bootstrapID);
  }
}
  
void Sun::sendConfigToSatellite(int masterSocket, int idx)
{
  assert((0<=idx) && (idx<numSatellites));

  /* Find out what the configuration should be */

  unsigned char pwd[20];
  int pwdLen;
  int sunInterval, planetInterval, planetPort;
  in_addr_t planetIp;
  getGoalSettingsForSatellite(idx, &sunInterval, &planetInterval, &planetIp, &planetPort, pwd, sizeof(pwd), &pwdLen);

  /* Generate a configuration packet */

  unsigned char data[24+pwdLen];
  data[0] = 0x53;
  data[1] = 0x41;
  data[2] = 0x54;
  data[3] = 0x4C;
  data[4] = 0x01;
  data[5] = 0x00;
  for (int i=0; i<8; i++)
    data[6+i] = satellite[idx].sunPassword[i];

  *(unsigned short*)&data[14] = htons(sunInterval);
  *(unsigned short*)&data[16] = htons(planetInterval);
  *(in_addr_t*)&data[18] = planetIp;
  *(unsigned short*)&data[22] = htons(planetPort);
  
  for (int i=0; i<pwdLen; i++)
    data[24+i] = pwd[i];
  
  /* Logging */
  
  char buf1[256], buf2[256], buf3[256];
  log->write(3, "Sending configuration to satellite %s:%s at %s:%d", satellite[idx].email, satellite[idx].identifier, ipdisp(satellite[idx].ip, buf1), satellite[idx].port);
  log->write(3, "Configuration is planet=%s:%d iSun=%d iPlanet=%d pSun=[%s] pPlanet=[%s]",
    ipdisp(planetIp, buf1), planetPort, sunInterval, planetInterval, renderBytes(satellite[idx].sunPassword, 8, buf2), renderBytes(pwd, pwdLen, buf3)
  );
  log->dump(4, data, sizeof(data));
  
  /* Send the configuration packet to the satellite */
  
  struct sockaddr_in dest;
  dest.sin_family = AF_INET;
  dest.sin_port = htons(satellite[idx].port);
  dest.sin_addr.s_addr = satellite[idx].ip;
  
  sendto(masterSocket, data, sizeof(data), MSG_DONTWAIT, (struct sockaddr*)&dest, sizeof(dest));
}

void Sun::tryToAllocateSatellites()
{
  /* This is called at various points when we think there is a chance to allocate new satellites
     to experiments, e.g., when a satellite has just come online, or when we've activated
     a new experiment. */

  bool progress = true;
  while (progress) {
    progress = false;
    
    /* Look for satellites that are not currently allocated */
    
    for (int i=0; i<numSatellites; i++) {
      if (!satellite[i].isAllocated) {
      
        /* If there are multiple experiments that still need satellites, we allocate the new satellite to the
           experiment with the lowest number of satellites (fairness). */
      
        int best = -1;
        for (int j=0; j<numExperiments; j++) {
          if ((experiment[j].numSatellitesMissing > 0) && (experiment[j].endTime >= satellite[i].nextRefreshAllocation))
            if ((best<0) || (experiment[best].numSatellitesAllocated > experiment[j].numSatellitesAllocated))
              best = j;
        }
        
        /* If an experiment was found, allocate the satellite to it */
        
        if (best >= 0) {
          progress = true;
          satellite[i].isAllocated = true;
          strncpy(satellite[i].expOwner, experiment[best].owner, sizeof(satellite[i].expOwner));
          strncpy(satellite[i].expIdentifier, experiment[best].identifier, sizeof(satellite[i].expIdentifier));
          satellite[i].planetPassword[0] = 0xAA;
          satellite[i].planetPasswordLen = 8;
          for (int j=1; j<satellite[i].planetPasswordLen; j++)
            satellite[i].planetPassword[j] = random()&0xFF;
          satellite[i].planetName[0] = 0;
          satellite[i].planetIP = 0;
          satellite[i].planetPort = 0;
          satellite[i].nextRefreshAllocation = experiment[best].endTime;
          experiment[best].numSatellitesMissing --;
          experiment[best].numSatellitesAllocated ++;
          
          char planetPassword[2*satellite[i].planetPasswordLen+100];
          planetPassword[0] = 0;
          for (int j=0; j<satellite[i].planetPasswordLen; j++)
            sprintf(&planetPassword[2*j], "%02X", satellite[i].planetPassword[j]);
          
          /* Record the allocation in the database, in case the sun is restarted */
          
          db->addSatellite(experiment[best].owner, experiment[best].identifier, satellite[i].email, satellite[i].identifier, planetPassword, NULL);
          
          /* Regenerate the matrix for this experiment */
          
          invalidateMatrix(experiment[best].owner, experiment[best].identifier);
          
          /* Start bootstrapping to find the planet that is closest to this satellite. This has to be 
             done every time a satellite joins a new experiment because different experiments could
             have different sets of planets */
          
          startBootstrappingSatellite(i);
        }
      }
    }
  }
}

void Sun::startBootstrappingSatellite(int idx)
{
  assert(satellite[idx].isAllocated && !satellite[idx].isBootstrapping);
  char buf[20000];
  char *names[MAX_PLANETS], *ips[MAX_PLANETS], *ports[MAX_PLANETS];
  
  /* If the experiment doesn't have any planets (yet), abort here. This is the reason why planets
     always need to be up before the start of the experiment. */
  
  int np = db->getPlanets(satellite[idx].expOwner, satellite[idx].expIdentifier, buf, sizeof(buf), MAX_PLANETS, names, ips, ports);
  if (np <= 0) {
    satellite[idx].planetIP = 0;
    satellite[idx].planetPort = 0;
    satellite[idx].planetName[0] = 0;
    log->write(1, "WARNING: Cannot start bootstrapping satellite '%s:%s'; cannot get planets", satellite[idx].email, satellite[idx].identifier);
    return;
  }
  
  /* Start bootstrapping */
  
  log->write(1, "Bootstrapping satellite '%s:%s' (experiment '%s:%s', %d planet(s))", satellite[idx].email, satellite[idx].identifier, satellite[idx].expOwner, satellite[idx].expIdentifier, np);
  
  satellite[idx].isBootstrapping = true;
  satellite[idx].bootstrapID = nextBootstrapID ++;
  satellite[idx].lastChangedBootstrap = time(0);
  satellite[idx].numBootstraps = np;
  satellite[idx].bootIP = (in_addr_t*) malloc(np * sizeof(in_addr_t));
  satellite[idx].bootPort = (int*) malloc(np * sizeof(int));
  satellite[idx].bootAddress = (char**) malloc(np * sizeof(char*));
  satellite[idx].bootAttempts = (int*) malloc(np * sizeof(int));
  satellite[idx].bootSuccesses = (int*) malloc(np * sizeof(int));
  satellite[idx].bootBestRTT = (int*) malloc(np * sizeof(int));
  satellite[idx].currentBootstrapIndex = 0;
  for (int i=0; i<np; i++) {
    satellite[idx].bootIP[i] = inet_addr(ips[i]);
    satellite[idx].bootPort[i] = atoi(ports[i]);
    satellite[idx].bootAddress[i] = (char*) malloc(strlen(names[i])+1);
    strcpy(satellite[idx].bootAddress[i], names[i]);
    satellite[idx].bootAttempts[i] = 0;
    satellite[idx].bootSuccesses[i] = 0;
    satellite[idx].bootBestRTT[i] = INFINITE_RTT;
  }
  
  if (np > 0) {
    log->write(2, "Next bootstrap is '%s:%d' (attempt %d/%d)", satellite[idx].bootAddress[0], satellite[idx].bootPort[0], 1, NUM_BOOTSTRAP_ATTEMPTS_PER_NODE);
    satellite[idx].bootAttempts[0] ++;
  }
}

void Sun::cancelBootstrappingSatellite(int idx)
{
  assert(satellite[idx].isBootstrapping);
  satellite[idx].isBootstrapping = false;
  
  /* Free all the data structures we've allocated for bootstrapping */
  
  if (satellite[idx].bootIP) {
    free(satellite[idx].bootIP);
    satellite[idx].bootIP = NULL;
  }
  if (satellite[idx].bootPort) {
    free(satellite[idx].bootPort);
    satellite[idx].bootPort = NULL;
  }
  if (satellite[idx].bootAddress) {
    for (int i=0; i<satellite[idx].numBootstraps; i++)
      free(satellite[idx].bootAddress[i]);
      
    free(satellite[idx].bootAddress);
    satellite[idx].bootAddress = NULL;
  }
  if (satellite[idx].bootAttempts) {
    free(satellite[idx].bootAttempts);
    satellite[idx].bootAttempts = NULL;
  }
  if (satellite[idx].bootSuccesses) {
    free(satellite[idx].bootSuccesses);
    satellite[idx].bootSuccesses = NULL;
  }
  if (satellite[idx].bootBestRTT) {
    free(satellite[idx].bootBestRTT);
    satellite[idx].bootBestRTT = NULL;
  }
}

void getDomain(in_addr_t ip, char *buf, int bufsize)
{
  struct hostent *he = gethostbyaddr((char*)&ip, sizeof(ip), AF_INET);
  if (he && he->h_name) {
    char domain[1000];
    strncpy(domain, he->h_name, sizeof(domain));
    domain[sizeof(domain)-1] = 0;
    int p = 0;
    while (domain[p])
      p++;
    while ((p>0) && (domain[p] != '.'))
      p--;
    if (p>0)
      p--;
    while ((p>0) && (domain[p] != '.'))
      p--;
    if (domain[p] == '.')
      p++;

    strncpy(buf, &domain[p], bufsize);
  } else {
    strncpy(buf, "unknown", bufsize);
  }
}

void Sun::flushSatelliteEntry(int idx)
{
  assert((0<=idx) && (idx<numSatellites));
  
  /* This is called when a satellite goes offline, or when its properties change significantly (different IP etc.) */
  
  log->write(3, "Satellite '%s:%s' was online for %d seconds, sent %lld bytes and received %lld bytes", satellite[idx].email, satellite[idx].identifier,
    satellite[idx].lastStatusReport-satellite[idx].lastWentOnline, satellite[idx].totalBytesSent, satellite[idx].totalBytesReceived
  );
  
  db->recordSatelliteCredits(satellite[idx].email, satellite[idx].identifier, satellite[idx].lastWentOnline, 
    satellite[idx].lastStatusReport-satellite[idx].lastWentOnline, satellite[idx].ip, satellite[idx].port,
    satellite[idx].domain, satellite[idx].version, satellite[idx].totalBytesSent, satellite[idx].totalBytesReceived
  );
  
  if (satellite[idx].isBootstrapping)
    cancelBootstrappingSatellite(idx);
    
  satellite[idx] = satellite[--numSatellites];
}

void Sun::handleStatusReport(int masterSocket, in_addr_t ip, int port, unsigned char *data, int length)
{
  char buf1[256], buf2[256];
  int now = time(0);

  /* Sanity checks */

  if (length < 33) {
    log->write(1, "WARNING: Truncated status report from %s:%d -- ignoring", ipdisp(ip, buf1), port);
    return;
  }
  
  if ((data[0] != 0x53) || (data[1] != 0x41) || (data[2] != 0x54) || (data[3] != 0x4C)) {
    log->write(1, "WARNING: Status report with incorrect magic number from %s:%d -- ignoring", ipdisp(ip, buf1), port);
    return;
  }
    
  /* Parse the packet */
    
  int firstZero = 32;
  while ((firstZero<length) && data[firstZero])
    firstZero ++;
  int secondZero = firstZero+1;
  while ((secondZero<length) && data[secondZero])
    secondZero ++;
  
  if (secondZero >= length) {
    log->write(1, "WARNING: Status report with broken identifier from %s:%d -- ignoring", ipdisp(ip, buf1), port);
    return;
  }
  
  char *email = (char*)&data[32];
  char *identifier = (char*)&data[firstZero+1];
  if (strlen(email)>MAX_EMAIL_LEN)
    email[MAX_EMAIL_LEN] = 0;
  if (strlen(identifier)>MAX_IDENTIFIER_LEN)
    identifier[MAX_IDENTIFIER_LEN] = 0;
  
  unsigned char *planetPassword = &data[secondZero+1];
  int planetPasswordLen = length - (secondZero+1);
  unsigned int reportedBytesSent = ntohl(*(unsigned int*)&data[24]);
  unsigned int reportedBytesReceived = ntohl(*(unsigned int*)&data[28]);
  int version = ntohs(*(unsigned short*)&data[4]);
  unsigned char *sunPassword = &data[6];
  
  log->write(3, "Status report from %s:%s at %s:%d", email, identifier, ipdisp(ip, buf1), port);
  log->dump(4, data, length);
  
  /* Find and update the satellite entry. If there isn't an entry for this satellite yet, create one. */
  
  int idx = -1;
  for (int i=0; i<numSatellites; i++) {
    if (!strcmp(satellite[i].email, email) && !strcmp(satellite[i].identifier, identifier))
      idx = i;
  }

  if ((idx>=0) && (satellite[idx].ip != ip)) {
    log->write(1, "Satellite '%s:%s' has migrated from %s:%d to %s:%d", satellite[idx].email, satellite[idx].identifier, ipdisp(satellite[idx].ip, buf1), satellite[idx].port, ipdisp(ip, buf2), port);
    flushSatelliteEntry(idx);
    idx = -1;
  }
  
  if ((idx>=0) && (satellite[idx].port != port)) {
    log->write(1, "WARNING: Two instances of satellite '%s:%s' appear to be active simultaneously (ports %d and %d on %s) -- ignoring port %d",
      satellite[idx].email, satellite[idx].identifier, satellite[idx].port, port, ipdisp(ip, buf1), port
    );
    return;
  }
  
  if (idx < 0) {
    if (numSatellites >= MAX_SATELLITES) {
      log->write(1, "WARNING: Too many satellites -- ignoring %s:%d", ipdisp(ip, buf1), port);
      return;
    }

    log->write(1, "Satellite '%s:%s' is now online at %s:%d", email, identifier, ipdisp(ip, buf1), port);
    
    idx = numSatellites ++;
    memset(&satellite[idx], 0, sizeof(satellite[idx]));
    strncpy(satellite[idx].email, email, sizeof(satellite[idx].email));
    strncpy(satellite[idx].identifier, identifier, sizeof(satellite[idx].identifier));
    getDomain(ip, satellite[idx].domain, sizeof(satellite[idx].domain));
    satellite[idx].reportedBytesSent = reportedBytesSent;
    satellite[idx].reportedBytesReceived = reportedBytesReceived;
    satellite[idx].ip = ip;
    satellite[idx].port = port;
    satellite[idx].version = version;
    memcpy(satellite[idx].sunPassword, sunPassword, 8);
    satellite[idx].lastWentOnline = now;
    satellite[idx].lastStatusReport = now;
    satellite[idx].totalBytesSent = 0;
    satellite[idx].totalBytesReceived = 0;
    satellite[idx].nextRefreshAllocation = now;
    satellite[idx].isBootstrapping = false;
  }
  
  /* We periodically check whether the satellite is still allocated to the same experiment.
     This is done infrequently, e.g. once per hour. */
  
  if (satellite[idx].nextRefreshAllocation <= now) {
    bool allowMigration = false;
    in_addr_t lastIP = 0;
    db->getSatelliteAllocation(email, identifier, now, &satellite[idx].isAllocated,
      satellite[idx].expOwner, satellite[idx].expIdentifier, satellite[idx].planetPassword, &satellite[idx].planetPasswordLen,
      satellite[idx].planetName, &satellite[idx].planetIP, &satellite[idx].planetPort, &lastIP, &allowMigration, &satellite[idx].nextRefreshAllocation
    );
    
    if (satellite[idx].isAllocated) {
      if (satellite[idx].planetIP) {
      
        /* It can happen that the satellite comes online with a different IP. Say a notebook is associated with planet A
           when the user is at work, then the user suspends the notebook and goes home. We can't simply associate the
           satellite with planet A again because another planet might be closer now. We check whether the satellite has
           appeared at this IP before. If it hasn't, we start bootstrapping again; otherwise we look up the planet it
           was associated with previously (say, planet B) and use that one. */
      
        if (allowMigration && (ip != lastIP)) {
          int validFrom;
          if (db->findBinding(satellite[idx].expOwner, satellite[idx].expIdentifier, email, identifier, ip, 
              satellite[idx].planetName, &satellite[idx].planetIP, &satellite[idx].planetPort, &validFrom)) {
            time_t ti = validFrom;
            log->write(2, "Reusing old binding of satellite '%s:%s' to planet '%s:%d' from %s", satellite[idx].email, 
              satellite[idx].identifier, satellite[idx].planetName, satellite[idx].planetPort, ctime(&ti)
            );
            db->addBinding(satellite[idx].expOwner, satellite[idx].expIdentifier, satellite[idx].email, satellite[idx].identifier, 
              ipdisp(satellite[idx].ip, buf1), satellite[idx].planetName, satellite[idx].planetPort
            );
            invalidateMatrix(satellite[idx].expOwner, satellite[idx].expIdentifier);
          } else {
            log->write(2, "Satellite '%s:%s' has never appeared at %s:%d before; bootstrapping",
              satellite[idx].email, satellite[idx].identifier, ipdisp(ip, buf1), port
            );
            
            satellite[idx].planetIP = 0;
            satellite[idx].planetPort = 0;
            satellite[idx].planetName[0] = 0;

            if (satellite[idx].isBootstrapping)
              cancelBootstrappingSatellite(idx);

            startBootstrappingSatellite(idx);
          }
        }
      } else {
        if (satellite[idx].isBootstrapping)
          cancelBootstrappingSatellite(idx);
          
        startBootstrappingSatellite(idx);
      }
      
      if (satellite[idx].planetIP) {
        log->write(1, "Satellite '%s:%s' is now allocated to experiment '%s:%s' (for another %d seconds, planet is %s:%d)",
          satellite[idx].email, satellite[idx].identifier, satellite[idx].expOwner, satellite[idx].expIdentifier,
          satellite[idx].nextRefreshAllocation-now, ipdisp(satellite[idx].planetIP, buf1), satellite[idx].port
        );
      } else {
        log->write(1, "Satellite '%s:%s' is now allocated to experiment '%s:%s' (for another %d seconds; bootstrapping)",
          satellite[idx].email, satellite[idx].identifier, satellite[idx].expOwner, satellite[idx].expIdentifier,
          satellite[idx].nextRefreshAllocation-now
        );
      }
    } else {
      log->write(1, "Satellite '%s:%s' is currently not allocated (for another %d seconds)", 
        satellite[idx].email, satellite[idx].identifier, satellite[idx].nextRefreshAllocation-now
      );
    }
  }

  /* Update various bookkeeping fields */

  unsigned int oldReportedBytesSent = satellite[idx].reportedBytesSent;
  unsigned int oldReportedBytesReceived = satellite[idx].reportedBytesReceived;
  satellite[idx].reportedBytesSent = reportedBytesSent;
  satellite[idx].reportedBytesReceived = reportedBytesReceived;
  
  if (satellite[idx].reportedBytesSent >= oldReportedBytesSent)
    satellite[idx].totalBytesSent += (satellite[idx].reportedBytesSent - oldReportedBytesSent);
  else if ((oldReportedBytesSent>=0xFC000000) && (satellite[idx].reportedBytesSent<0x04000000))
    satellite[idx].totalBytesSent += (4294967296LL-oldReportedBytesSent)+satellite[idx].reportedBytesSent;

  if (satellite[idx].reportedBytesReceived >= oldReportedBytesReceived)
    satellite[idx].totalBytesReceived += (satellite[idx].reportedBytesReceived - oldReportedBytesReceived);
  else if ((oldReportedBytesReceived>=0xFC000000) && (satellite[idx].reportedBytesReceived<0x04000000))
    satellite[idx].totalBytesReceived += (4294967296LL-oldReportedBytesReceived)+satellite[idx].reportedBytesReceived;

  memcpy(satellite[idx].sunPassword, sunPassword, sizeof(satellite[idx].sunPassword));
  satellite[idx].lastStatusReport = time(0);
  
  /* Handle timeouts during the bootstrap phase. Normally, when we tell the satellite to measure the RTT to some planet X,
     we expect to receive a RTT report from planet X soon afterwards. But sometimes that doesn't happen, e.g., when the
     report is lost, or when planet X is offline. In this case, we tell the satellite to simply try the next planet. */
  
  if (satellite[idx].isBootstrapping && (satellite[idx].lastChangedBootstrap<=(now-BOOTSTRAP_TIMEOUT_SEC_PER_ATTEMPT))) {
    log->write(2, "Satellite '%s:%s' reports bootstrap timeout for planet '%s:%d' (attempt %d/%d)",
      satellite[idx].email, satellite[idx].identifier, satellite[idx].bootAddress[satellite[idx].currentBootstrapIndex], 
      satellite[idx].bootPort[satellite[idx].currentBootstrapIndex], satellite[idx].bootAttempts[satellite[idx].currentBootstrapIndex],
      NUM_BOOTSTRAP_ATTEMPTS_PER_NODE
    );
    
    int nb = getBestNextBootstrap(idx);
    if (nb >= 0) {
      log->write(2, "Next bootstrap is '%s:%d' (attempt %d/%d)", satellite[idx].bootAddress[nb], satellite[idx].bootPort[nb], 1+satellite[idx].bootAttempts[nb], NUM_BOOTSTRAP_ATTEMPTS_PER_NODE);
      satellite[idx].currentBootstrapIndex = nb;
      satellite[idx].bootstrapID = nextBootstrapID ++;
      satellite[idx].lastChangedBootstrap = time(0);
      satellite[idx].bootAttempts[nb] ++;
    } else {
      finishBootstrappingSatellite(idx);
    }
  }

  /* If the satellite is not allocated, check whether there are any experiments that still need satellites */
  
  tryToAllocateSatellites();

  /* Send a configuration packet, if necessary */
  
  int sunIntervalIs = ntohs(*(unsigned short*)&data[14]);
  int planetIntervalIs = ntohs(*(unsigned short*)&data[16]);
  in_addr_t planetIpIs = *(in_addr_t*)&data[18];
  int planetPortIs = ntohs(*(unsigned short*)&data[22]);
  
  int sunIntervalShouldBe, planetIntervalShouldBe, planetPortShouldBe, pwdLenShouldBe;
  unsigned char pwdShouldBe[20];
  in_addr_t planetIpShouldBe;

  getGoalSettingsForSatellite(idx, &sunIntervalShouldBe, &planetIntervalShouldBe, &planetIpShouldBe, &planetPortShouldBe, pwdShouldBe, sizeof(pwdShouldBe), &pwdLenShouldBe);

  bool needToRespond = false;
  if (sunIntervalIs != sunIntervalShouldBe) 
    needToRespond = true;
  if (planetIntervalIs != planetIntervalShouldBe) 
    needToRespond = true;
  if (planetIpIs != planetIpShouldBe) 
    needToRespond = true;
  if (planetPortIs != planetPortShouldBe) 
    needToRespond = true;
  
  if (planetPasswordLen != pwdLenShouldBe) {
    needToRespond = true;
  } else {
    for (int i=0; i<pwdLenShouldBe; i++) {
      if (pwdShouldBe[i] != planetPassword[i]) 
        needToRespond = true;
    }
  }
  
  if (needToRespond)
    sendConfigToSatellite(masterSocket, idx);
}

void Sun::handleRTTReport(int masterSocket, in_addr_t ip, int port, int rttMillis, unsigned char *planetPassword, int planetPasswordLen)
{
  char buf1[256], buf2[256];
  log->write(3, "RTT report from %s:%d (rtt=%dms, pwd=[%s])", ipdisp(ip, buf1), port, rttMillis, renderBytes(planetPassword, planetPasswordLen, buf2));

  /* The process is:
       1) The satellite joins
       2) The sun tells the satellite to associate with the first planet, using an ID of 0xBB....
       3) The satellite sends its first status report to this planet
       4) The planet sees that the ID starts with 0xBB, so it sends a single probe packet
       5) The satellite responds to the probe
       6) The planet measures the RTT and sends an RTT report to the sun
     At this point, this method is called. We begin with some sanity checks. */

  if ((planetPasswordLen != 5) || (planetPassword[0] != 0xBB)) {
    log->write(1, "WARNING: Invalid RTT report from %s:%d -- ignoring", ipdisp(ip, buf1), port);
    return;
  }

  /* Extract the bootstrap ID (it uniquely identifies a satellite during the bootstrap process)
     and look for the satellite in our local data structures. */
  
  int bootstrapID = ntohl(*(unsigned int*)&planetPassword[1]);

  int idx = -1;
  for (int i=0; i<numSatellites; i++) {
    if (satellite[i].isBootstrapping && (satellite[i].bootstrapID == bootstrapID)) 
      idx = i;
  }
  
  if (idx < 0) {
    log->write(1, "WARNING: Spurious RTT report from %s:%d (BootID %08X) -- ignoring", ipdisp(ip, buf1), port, bootstrapID);
    return;
  }
  
  /* Sometimes RTT reports might be delayed for a long time, so we may have switched to the
     next planet already. In this case, we drop the RTT report (shouldn't happen often). */
  
  assert((0<=satellite[idx].currentBootstrapIndex) && (satellite[idx].currentBootstrapIndex<satellite[idx].numBootstraps));
  if ((satellite[idx].bootIP[satellite[idx].currentBootstrapIndex] != ip) || (satellite[idx].bootPort[satellite[idx].currentBootstrapIndex] != port)) {
    log->write(1, "WARNING: RTT report from %s:%d, but satellite's current bootstrap is %s:%d -- ignoring",
      ipdisp(ip, buf1), port, ipdisp(satellite[idx].bootIP[satellite[idx].currentBootstrapIndex], buf2), satellite[idx].bootPort[satellite[idx].currentBootstrapIndex]
    );
    return;
  }
  
  /* Record the RTT for this planet... */
  
  log->write(2, "Measured RTT from %s:%s to %s:%d is %dms (attempt %d/%d)", satellite[idx].email, satellite[idx].identifier, 
    satellite[idx].bootAddress[satellite[idx].currentBootstrapIndex], satellite[idx].bootPort[satellite[idx].currentBootstrapIndex], rttMillis,
    satellite[idx].bootAttempts[satellite[idx].currentBootstrapIndex], NUM_BOOTSTRAP_ATTEMPTS_PER_NODE
  );
  
  satellite[idx].bootSuccesses[satellite[idx].currentBootstrapIndex] ++;
  if (rttMillis < satellite[idx].bootBestRTT[satellite[idx].currentBootstrapIndex])
    satellite[idx].bootBestRTT[satellite[idx].currentBootstrapIndex] = rttMillis;
    
  /* ... and try the next planet */
    
  int nb = getBestNextBootstrap(idx);
  if (nb >= 0) {
    log->write(2, "Next bootstrap is '%s:%d' (attempt %d/%d)", satellite[idx].bootAddress[nb], satellite[idx].bootPort[nb], 1+satellite[idx].bootAttempts[nb], NUM_BOOTSTRAP_ATTEMPTS_PER_NODE);
    satellite[idx].currentBootstrapIndex = nb;
    satellite[idx].bootstrapID = nextBootstrapID ++;
    satellite[idx].lastChangedBootstrap = time(0);
    satellite[idx].bootAttempts[nb] ++;
  } else {
    finishBootstrappingSatellite(idx);
  }
}

int Sun::findExperimentEntry(const char *owner, const char *identifier)
{
  for (int i=0; i<numExperiments; i++) {
    if (!strcmp(experiment[i].owner, owner) && !strcmp(experiment[i].identifier, identifier)) 
      return i;
  }
  
  return -1;
}

void Sun::invalidateMatrix(const char *owner, const char *identifier)
{
  log->write(3, "Invalidating matrix for experiment '%s:%s'", owner, identifier);
  
  for (int i=0; i<numMatrixCacheEntries; i++) {
    if (!strcmp(matrixCache[i].owner, owner) && !strcmp(matrixCache[i].identifier, identifier)) {
      if (matrixCache[i].matrix)
        free(matrixCache[i].matrix);
        
      matrixCache[i] = matrixCache[--numMatrixCacheEntries];
      break;
    }
  }
}

int Sun::findInMatrixCache(const char *owner, const char *identifier)
{
  for (int i=0; i<numMatrixCacheEntries; i++) {
    if (!strcmp(matrixCache[i].owner, owner) && !strcmp(matrixCache[i].identifier, identifier)) {
      matrixCache[i].lastQueried = time(0);
      return i;
    }
  }
  
  while (numMatrixCacheEntries >= MAX_MATRIX_CACHE_ENTRIES) {
    int oldest = 0;
    for (int i=1; i<numMatrixCacheEntries; i++) {
      if (matrixCache[oldest].lastQueried > matrixCache[i].lastQueried)
        oldest = i;
    }
    
    if (matrixCache[oldest].matrix)
      free(matrixCache[oldest].matrix);
      
    matrixCache[oldest] = matrixCache[--numMatrixCacheEntries];
  }
    
  /* Entry not found -- must regenerate */
    
  log->write(3, "Regenerating matrix for experiment '%s:%s'", owner, identifier);
    
  int numSatellites = db->getNumSatellitesInExperiment(owner, identifier);
  if (numSatellites < 0)
    return -EDATABASE;

  bool dynAllocSats;
  int satsLeftToAllocate, endTime;
  if (!db->statExperiment(owner, identifier, NULL, &endTime, NULL, &dynAllocSats, NULL, &satsLeftToAllocate, NULL, 0))
    return -ENOEXP;
    
  if (endTime < time(0)) {
    log->write(1, "WARNING: Matrix requested for old experiment '%s:%s' -- ignoring", owner, identifier);
    return -EEXPIRED;
  }
    
  if (dynAllocSats)
    numSatellites += satsLeftToAllocate;
    
  int idx = numMatrixCacheEntries++;
  strncpy(matrixCache[idx].owner, owner, sizeof(matrixCache[idx].owner));
  strncpy(matrixCache[idx].identifier, identifier, sizeof(matrixCache[idx].identifier));
  matrixCache[idx].lastQueried = time(0);
  matrixCache[idx].matrixSizeBytes = numSatellites * MATRIX_ENTRY_BYTES;
  matrixCache[idx].matrix = (unsigned char*) malloc(matrixCache[idx].matrixSizeBytes);
  memset(matrixCache[idx].matrix, 0, matrixCache[idx].matrixSizeBytes);
  db->getMatrixForExperiment(owner, identifier, matrixCache[idx].matrix, numSatellites);
  log->dump(4, matrixCache[idx].matrix, matrixCache[idx].matrixSizeBytes);

  /* Now calculate a SHA1 hash */

  SHA_CTX c;
  SHA1_Init(&c);
  if (matrixCache[idx].matrixSizeBytes > 0)
    SHA1_Update(&c, matrixCache[idx].matrix, matrixCache[idx].matrixSizeBytes);
  SHA1_Final(matrixCache[idx].matrixHash, &c);

  char buf1[256];
  log->write(3, "Hash is [%s]", renderBytes(matrixCache[idx].matrixHash, SHA_DIGEST_LENGTH, buf1));

  return idx;
}

void Sun::timerExpired(int id, int arg)
{
  int now = time(0);
  
  /* We periodically remove the entries of planets and satellites that have gone offline */
  
  if (id == TI_HOUSEKEEPING) {
    for (int i=0; i<numSatellites; ) {
      if (satellite[i].lastStatusReport <= (now-SATELLITE_CLEANUP_SEC)) {
        log->write(3, "Removing entry for satellite '%s:%s'", satellite[i].email, satellite[i].identifier);
        flushSatelliteEntry(i);
      } else {
        i++;
      }
    }
    
    for (int i=0; i<numPlanets; ) {
      if (planet[i].lastStatusReport <= (now-PLANET_CLEANUP_SEC)) {
        log->write(3, "Removing entry for planet '%s:%d' in '%s:%s'", planet[i].address, planet[i].port, planet[i].expOwner, planet[i].expIdentifier);
        flushPlanetEntry(i);
      } else {
        i++;
      }
    }

    selector->setTimer(TI_HOUSEKEEPING, 0, selector->getTimeMicros()+HOUSEKEEPING_INTERVAL_MICROS, this);
    return;
  }
  
  /* We periodically record statistics about each satellite in the database. This can be used
     when allocating satellites to experiments; for example, we might give more satellites to
     experimenters who have contributed more */
  
  if (id == TI_FLUSH_CREDIT) {
    log->write(2, "Recording credit in the database");
    for (int i=0; i<numSatellites; i++) {
      if (satellite[i].lastWentOnline < satellite[i].lastStatusReport) {
        log->write(3, "Satellite '%s:%s' was online for %d seconds, sent %lld bytes and received %lld bytes", satellite[i].email, satellite[i].identifier,
          satellite[i].lastStatusReport-satellite[i].lastWentOnline, satellite[i].totalBytesSent, satellite[i].totalBytesReceived
        );
        db->recordSatelliteCredits(satellite[i].email, satellite[i].identifier, satellite[i].lastWentOnline, 
          satellite[i].lastStatusReport-satellite[i].lastWentOnline, satellite[i].ip, satellite[i].port,
          satellite[i].domain, satellite[i].version, satellite[i].totalBytesSent, satellite[i].totalBytesReceived
        );
        
        satellite[i].lastWentOnline = satellite[i].lastStatusReport;
        satellite[i].totalBytesSent = 0;
        satellite[i].totalBytesReceived = 0;
      }
    }
  
    selector->setTimer(TI_FLUSH_CREDIT, 0, selector->getTimeMicros()+FLUSH_CREDIT_INTERVAL_MICROS, this);
    return;
  }
  
  /* We periodically reload the current set of experiments from the database (once every hour).
     Thus, we don't have to keep all future experiments in memory. */
  
  if (id == TI_EXPERIMENT_REFRESH) {
    char buf[200*MAX_EXPERIMENTS];
    char *owners[MAX_EXPERIMENTS];
    char *identifiers[MAX_EXPERIMENTS];
    numExperiments = db->getExperiments(now, now+3600, buf, sizeof(buf), MAX_EXPERIMENTS, owners, identifiers);
    if (numExperiments >= 0) {
      for (int i=0; i<numExperiments; i++) {
        memset(&experiment[i], 0, sizeof(experiment[i]));
        strncpy(experiment[i].owner, owners[i], sizeof(experiment[i].owner));
        strncpy(experiment[i].identifier, identifiers[i], sizeof(experiment[i].identifier));
        experiment[i].numSatellitesAllocated = db->getNumSatellitesInExperiment(owners[i], identifiers[i]);
        experiment[i].numPlanetsAllocated = db->getNumPlanetsInExperiment(owners[i], identifiers[i]);
        if (!db->statExperiment(owners[i], identifiers[i], &experiment[i].startTime, &experiment[i].endTime, &experiment[i].dynAllocPlanets, 
            &experiment[i].dynAllocSats, &experiment[i].allowSatMigration, &experiment[i].numSatellitesMissing, 
            experiment[i].description, sizeof(experiment[i].description))) {
          log->write(1, "WARNING: Cannot stat experiment '%s:%s'; dynamic satellite allocation disabled", owners[i], identifiers[i]);
        }
      }
    } else {
      log->write(1, "WARNING: Cannot refresh list of experiments; dynamic satellite allocation may not work");
      numExperiments = 0;
    }
    
    tryToAllocateSatellites();
  
    selector->setTimer(TI_EXPERIMENT_REFRESH, 0, selector->getTimeMicros()+EXPERIMENT_REFRESH_INTERVAL_MICROS, this);
    return;
  }
  
  /* This should never be reached */
  
  assert(false);
}

void Sun::flushPlanetEntry(int idx)
{
  planet[idx] = planet[--numPlanets];
}

bool Sun::isPlanetConfiguredCorrectly(int idx)
{
  if ((idx<0) || (idx>=numPlanets))
    return false;
    
  int midx = findInMatrixCache(planet[idx].expOwner, planet[idx].expIdentifier);
  if (midx < 0)
    return false;
    
  if (memcmp(matrixCache[midx].matrixHash, planet[idx].lastConfigHash, SHA_DIGEST_LENGTH))
    return false;
    
  return true;
}

void Sun::handlePlanetReport(int masterSocket, in_addr_t ip, int port, unsigned char *data, int length)
{
  char buf1[256];
  int now = time(0);
  data[length] = 0;

  /* Sanity checks */

  if (length < 8+SHA_DIGEST_LENGTH) {
    log->write(1, "WARNING: Truncated planet report from %s:%d -- ignoring", ipdisp(ip, buf1), port);
    return;
  }
  
  if ((data[0] != 0x53) || (data[1] != 0x55) || (data[2] != 0x4E) || (data[3] != 0x52)) {
    log->write(1, "WARNING: Planet report with incorrect magic number from %s:%d -- ignoring", ipdisp(ip, buf1), port);
    return;
  }
    
  int firstZero = 7+SHA_DIGEST_LENGTH;
  while ((firstZero<length) && data[firstZero])
    firstZero ++;
  
  if (firstZero >= length) {
    log->write(1, "WARNING: Planet report with broken identifier from %s:%d -- ignoring", ipdisp(ip, buf1), port);
    return;
  }
  
  /* Decode packet. We get:
       - the owner of the experiment,
       - the identifier of the experiment,
       - the version of the planet proxy,
       - a flag that indicates whether or not the proxy is configured to accept commands from the sun, and
       - the hash of the planet's current configuration */
  
  char *owner = (char*)&data[7+SHA_DIGEST_LENGTH];
  char *identifier = (char*)&data[firstZero+1];
  if (strlen(owner)>=199)
    owner[199] = 0;
  if (strlen(identifier)>=199)
    identifier[199] = 0;
    
  int version = ntohs(*(unsigned short*)&data[4]);
  bool acceptsUpdates = (data[6]!=0);
  unsigned char *configHash = &data[7];
  
  log->write(3, "Planet report from %s:%d (experiment %s:%s)", ipdisp(ip, buf1), port, owner, identifier);
  log->dump(4, data, length);

  /* Do we already have an entry for this planet? */

  int idx = -1;
  for (int i=0; i<numPlanets; i++) {
    if ((planet[i].ip == ip) && (planet[i].port == port))
      idx = i;
  }
  
  if ((idx>=0) && (strcmp(identifier, planet[idx].expIdentifier) || strcmp(owner, planet[idx].expOwner))) {
    log->write(1, "WARNING: Planet %s:%d has changed from experiment '%s:%s' to '%s:%s'", planet[idx].address, port, planet[idx].expOwner, planet[idx].expIdentifier, owner, identifier);
    flushPlanetEntry(idx);
    idx = -1;
  }
  
  /* If we don't have an entry (yet?), the next steps depend on whether the experiment is configured
     to use dynamic planets. If this is not the case, the planet doesn't belong to the experiment,
     and we must ignore it. */
  
  if (idx < 0) {
    if (numPlanets >= MAX_PLANETS) {
      log->write(1, "WARNING: Too many planets -- ignoring report from %s:%d", ipdisp(ip, buf1), port);
      return;
    }
    
    char planetAddress[200];
    if (!db->statPlanet(owner, identifier, ip, port, planetAddress)) {
      int startTime;
      bool dynAllocPlanets;
      if (!db->statExperiment(owner, identifier, &startTime, NULL, &dynAllocPlanets, NULL, NULL, NULL, NULL, 0)) {
        log->write(1, "WARNING: Planet report from %s:%d for unknown experiment '%s:%s' -- ignoring", ipdisp(ip, buf1), port, owner, identifier);
        return;
      }
      
      /* If the dynamic planets feature is enabled, the user hasn't specified a list of planets
         on the web interface; we just add the planets to the database as they come online. */
      
      if (!dynAllocPlanets) {
        log->write(1, "WARNING: Planet report for static experiment '%s:%s' from unknown planet %s:%d -- ignoring", owner, identifier, ipdisp(ip, buf1), port);
        return;
      }

      /* OK, let's admit the new planet and record it in the database */

      struct hostent *he = gethostbyaddr((char*)&ip, sizeof(ip), AF_INET);
      if (he && he->h_name) 
        strncpy(planetAddress, he->h_name, sizeof(planetAddress));
      else
        ipdisp(ip, planetAddress);
      
      db->addPlanet(owner, identifier, planetAddress, ip, port);
      invalidateMatrix(owner, identifier);
      
      int iexp = findExperimentEntry(owner, identifier);
      if (iexp >= 0)
        experiment[iexp].numPlanetsAllocated ++;
    }
    
    idx = numPlanets ++;
    planet[idx].ip = ip;
    planet[idx].port = port;
    planet[idx].version = version;
    planet[idx].acceptsUpdates = acceptsUpdates;
    strncpy(planet[idx].expOwner, owner, sizeof(planet[idx].expOwner));
    strncpy(planet[idx].expIdentifier, identifier, sizeof(planet[idx].expIdentifier));
    strncpy(planet[idx].address, planetAddress, sizeof(planet[idx].address));
  }
  
  memcpy(planet[idx].lastConfigHash, configHash, SHA_DIGEST_LENGTH);
  planet[idx].lastStatusReport = now;
  log->write(1, "Planet report from %s:%d (experiment %s:%s)", planet[idx].address, port, owner, identifier);
  
  /* Check whether the planet is using the latest configuration. The sun changes the configuration occasionally,
     e.g. when a satellite associates with a new planet. We check the SHA-1 hash that the planet has sent
     against the SHA-1 hash of our current configuration. If they don't match, we send an update to the planet. */
  
  if (planet[idx].acceptsUpdates) {
    int midx = findInMatrixCache(owner, identifier);
    
    /* Sometimes we _cannot_ send a configuration, e.g. when the planet is part of an experiment that has
       already terminated. In this case, we send an error message to the planet, which is added to the planet's log.
       Thus, the user can see what the problem is. */
    
    if (midx < 0) {
      log->write(1, "WARNING: Cannot generate matrix for experiment '%s:%s' (code %d)", owner, identifier, -midx);
      unsigned char packet[500];
      *(unsigned int*)&packet[0] = htonl(0x52455252);
      *(unsigned short*)&packet[4] = htons(0x0100);
      *(unsigned short*)&packet[6] = htons(-midx);
      packet[8] = 0;
      if (midx == -EDATABASE)
        strcpy((char*)&packet[8], "Database error");
      else if (midx == -ENOEXP)
        snprintf((char*)&packet[8], sizeof(packet)-8, "No such experiment: %s/%s", owner, identifier);
      else if (midx == -EEXPIRED)
        snprintf((char*)&packet[8], sizeof(packet)-8, "Experiment has expired: %s/%s", owner, identifier);
      else
        snprintf((char*)&packet[8], sizeof(packet)-8, "Unknown error: %d", -midx);
        
      struct sockaddr_in dest;
      dest.sin_family = AF_INET;
      dest.sin_port = htons(planet[idx].port);
      dest.sin_addr.s_addr = planet[idx].ip;

      log->dump(4, packet, sizeof(packet));
      sendto(masterSocket, packet, 8+strlen((char*)&packet[8]), MSG_DONTWAIT, (struct sockaddr*)&dest, sizeof(dest));      
      return;
    }
    
    /* We do have a configuration; send it if the planet's configuration is not up to date. */
    
    if (memcmp(planet[idx].lastConfigHash, matrixCache[midx].matrixHash, SHA_DIGEST_LENGTH)) {
      log->write(3, "Planet config hash is [%s]", renderBytes(planet[idx].lastConfigHash, SHA_DIGEST_LENGTH, buf1));
      log->write(3, "Latest config hash is [%s]", renderBytes(matrixCache[midx].matrixHash, SHA_DIGEST_LENGTH, buf1));
      log->write(1, "Sending new matrix to planet %s:%d", planet[idx].address, planet[idx].port);
      assert(matrixCache[midx].matrixSizeBytes < 65529);
      unsigned char packet[matrixCache[midx].matrixSizeBytes+6];
      *(unsigned int*)&packet[0] = htonl(0x434f4e46);
      *(unsigned short*)&packet[4] = htons(0x0100);
      memcpy(&packet[6], matrixCache[midx].matrix, matrixCache[midx].matrixSizeBytes);

      struct sockaddr_in dest;
      dest.sin_family = AF_INET;
      dest.sin_port = htons(planet[idx].port);
      dest.sin_addr.s_addr = planet[idx].ip;

      log->dump(4, packet, sizeof(packet));
      sendto(masterSocket, packet, sizeof(packet), MSG_DONTWAIT, (struct sockaddr*)&dest, sizeof(dest));
    }
  }
}

void Sun::descriptorReady(int fd, bool canRead, bool canWrite, bool hasException)
{
  assert((fd == masterSocket) && canRead);

  unsigned char packet[65536];
  struct sockaddr_in addr;
  socklen_t fromlen = sizeof(addr);
  int size = recvfrom(masterSocket, packet, sizeof(packet), 0, (struct sockaddr*)&addr, &fromlen);
  if (size <= 0)
    panic("Cannot read from UDP master port (errno=%d)", errno);

  /* The sun can receive three types of packets:
       1) Satellite status reports
       2) Planet status reports, and 
       3) RTT reports (from planets) 
     We use the magic number in the first four bytes to distinguish between them. */

  if ((size >= 4) && (ntohl(*(unsigned int*)&packet[0]) == 0x5341544C))
    handleStatusReport(masterSocket, addr.sin_addr.s_addr, ntohs(addr.sin_port), packet, size);
  else if ((size > 6) && (ntohl(*(unsigned int*)&packet[0]) == 0x4D525454))
    handleRTTReport(masterSocket, addr.sin_addr.s_addr, ntohs(addr.sin_port), ntohs(*(unsigned short*)&packet[4]), &packet[6], size - 6);
  else if ((size >=4 ) && (ntohl(*(unsigned int*)&packet[0]) == 0x53554e52))
    handlePlanetReport(masterSocket, addr.sin_addr.s_addr, ntohs(addr.sin_port), packet, size);
  else
    log->write(1, "WARNING: Unknown packet type");
}

bool Sun::isValidAuthentication(const char *owner, const char *auth)
{
  assert(owner && auth);
  char username[200], *saveptr;
  strncpy(username, auth, sizeof(username));
  strtok_r(username, ":", &saveptr);
  char actualOwner[200], actualPassword[200];
  if (!db->statUser(username, actualOwner, sizeof(actualOwner), actualPassword, sizeof(actualPassword)))
    return false;
  if (strcmp(owner, actualOwner))
    return false;
    
  char actualAuth[401];
  snprintf(actualAuth, sizeof(actualAuth), "%s:%s", username, actualPassword);
  return !strcmp(auth, actualAuth);
}

void Sun::reassociateSatellites(const char *owner, const char *identifier)
{
  log->write(1, "Reassociating satellites for experiment %s:%s", owner, identifier);
  for (int i=0; i<numSatellites; i++) {
    if (satellite[i].isAllocated && !strcmp(satellite[i].expOwner, owner) && !strcmp(satellite[i].expIdentifier, identifier)) {
      if (satellite[i].isBootstrapping)
        cancelBootstrappingSatellite(i);

      startBootstrappingSatellite(i);
    }
  }
}
